import os
import platform
import shutil
from typing import Any

from sqlalchemy import (
    create_engine, MetaData, Table, Column, String, select, update, inspect
)
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError

from NepTrainKit import module_path, get_user_config_path


class Config:
    """
使用数据库保存软件配置
    """
    _instance = None
    init_flag = False

    def __new__(cls, *args):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        if Config.init_flag:
            return
        Config.init_flag = True
        self.connect_db()

    def connect_db(self):
        user_config_path = get_user_config_path()

        db_file = os.path.join(user_config_path, "config.sqlite")
        if not os.path.exists(db_file):
            if not os.path.exists(user_config_path):
                os.makedirs(user_config_path)
            shutil.copy(os.path.join(module_path, 'Config/config.sqlite'), db_file)

        # Initialize SQLAlchemy engine for SQLite
        url = f"sqlite:///{db_file}"
        # check_same_thread=False to be safe with GUI contexts
        self.engine: Engine = create_engine(url, future=True)

        # Ensure the table exists (reflect if present, otherwise create)
        self._metadata = MetaData()
        inspector = inspect(self.engine)
        if inspector.has_table('config'):
            self._config_table = Table('config', self._metadata, autoload_with=self.engine)
        else:
            self._config_table = Table(
                'config', self._metadata,
                Column('section', String, primary_key=True),
                Column('option', String, primary_key=True),
                Column('value', String)
            )
            self._metadata.create_all(self.engine)

    @classmethod
    def get_path(cls,section="setting", option="last_path")->str:
        """
        获取上一次文件交互的路径
        :param section:
        :param option:
        :return:
        """
        path = cls.get(section, option)
        if path:
            if os.path.exists(path):
                return path
        return "./"

    @classmethod
    def has_option(cls,section, option) ->bool:
        if cls.get(section,option) is not None:
            return True
        return False

    @classmethod
    def getboolean(cls, section, option, fallback=None)->bool|None:
        v = cls.get(section, option,fallback)
        try:
            v = eval(v)
        except:
            v = None
        if v is None:
            return fallback
        return v

    @classmethod
    def getint(cls, section, option, fallback=None) ->int|None:
        v = cls.get(section, option,fallback)

        try:
            v = int(v)
        except:

            v = None
        if v is None:
            return fallback

        return v
    @classmethod
    def getfloat(cls,section,option,fallback=None)->float|None:
        v=    cls.get(section,option,fallback)

        try:
            v=float(v)
        except:

            v=None
        if v is None:
            return fallback
        return v
    @classmethod
    def get(cls,section,option,fallback=None)->Any:
        try:
            cfg = cls._instance
            table = cfg._config_table
            with cfg.engine.begin() as conn:
                stmt = select(table.c.value).where(
                    table.c.section == section,
                    table.c.option == option
                ).limit(1)
                result = conn.execute(stmt).scalar_one_or_none()
            if result is None:
                if fallback is not None:
                    cls.set(section, option, fallback)
                return fallback
            return result
        except SQLAlchemyError:
            # Fallback behavior in case of unexpected DB errors
            return fallback

    @classmethod
    def set(cls,section,option,value):
        if option == "theme":
            cls.theme = value
        cfg = cls._instance
        table = cfg._config_table
        val_str = str(value)
        with cfg.engine.begin() as conn:
            # Try update first; if no row affected, insert
            upd = (
                update(table)
                .where(table.c.section == section, table.c.option == option)
                .values(value=val_str)
            )
            res = conn.execute(upd)
            if res.rowcount == 0:
                ins = table.insert().values(section=section, option=option, value=val_str)
                conn.execute(ins)

    @classmethod
    def update_section(cls,old,new):
        cfg = cls._instance
        table = cfg._config_table
        with cfg.engine.begin() as conn:
            stmt = update(table).where(table.c.section == old).values(section=new)
            conn.execute(stmt)
Config()



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/18 17:14
# @Author  : 兵
# @email    : 1747193328@qq.com
import os
import re
import traceback
from functools import partial
from pathlib import Path
import numpy as np
import numpy.typing as npt
from loguru import logger
from NepTrainKit.core import MessageManager


def get_rmse(array1, array2):
    return np.sqrt(((array1 - array2) ** 2).mean())

def read_nep_in(file_name: str|Path) ->dict[str,str]:
    """
    将nep.in的内容解析成dict
    :param file_name: nep.in的路径
    :return:
    """
    run_in={}

    if  not os.path.exists(file_name):
        return run_in
    try:
        with open(file_name, 'r', encoding="utf8") as f:

            groups = re.findall(r"^([A-Za-z_]+)\s+([^\#\n]*)", f.read(), re.MULTILINE)

            for group in groups:

                run_in[group[0].strip()] = group[1].strip()
    except:
        logger.debug(traceback.format_exc())
        MessageManager.send_warning_message("read nep.in file error")
        nep_in = {}
    return run_in

def check_fullbatch(run_in:dict[str,str],structure_num:int)->bool:

    if run_in.get("prediction")=="1":
        return True
    if int(run_in.get("batch",1000))>=structure_num:
        return True
    return False

def read_nep_out_file(file_path:Path|str,**kwargs)->npt.NDArray[np.float32]:
    """
    读取out数值文件
    :param file_path: energy_train.out
    :param kwargs:
    :return:
    """
    if os.path.exists(file_path):

        data = np.loadtxt(file_path,**kwargs)
        logger.info("Reading file: {},shape:{}".format(file_path,data.shape))

        return data
    else:
        return np.array([])

def parse_array_by_atomnum(array: npt.NDArray[np.float32],
                           atoms_num_list: npt.NDArray[np.float32],
                           map_func=np.linalg.norm,
                           axis:int=0
                           )->npt.NDArray[np.float32]:
    """
    根据一个映射列表，将原数组按照原子数列表拆分，
    这个主要是处理文件中原子数不一致的情况，比如力 描述符等文件是按照原子数的 把他们转换成结构的
    :param array: 原数组
    :param atoms_num_list: 原子数列表
    :param map_func: 需要对每个结构的数据进行处理的函数 比如求平均 求和等
    :param axis: 映射轴
    :return: 映射后的数组

    """
    if len(array)==0:
        return array
    # 使用 np.cumsum() 计算每个分组的结束索引
    split_indices = np.cumsum(atoms_num_list)[:-1]
    # 使用 np.split() 按照分组拆分数组
    split_arrays = np.split(array, split_indices)
    func = partial(map_func, axis=axis)

    # 对每个分组求和，使用 np.vectorize 进行向量化
    new_array = np.array(list(map(func, split_arrays)))
    return new_array

def get_nep_type(file_path:Path|str)->int:
    """
    根据nep.txt 判断势函数类别
    """
    nep_type_to_model_type = {
        "nep3": 0,
        "nep3_zbl": 0,
        "nep3_dipole": 1,
        "nep3_polarizability": 2,
        "nep4": 0,
        "nep4_zbl": 0,
        "nep4_dipole": 1,
        "nep4_polarizability": 2,
        "nep4_zbl_temperature":3,
        "nep4_temperature":3,
        "nep5": 0,
        "nep5_zbl": 0
    }
    model_type=0
    try:
        with open(file_path, 'r') as file:
            # 读取第一行
            first_line = file.readline().strip()
            parts = first_line.split()
            nep_type=parts[0]
            model_type = nep_type_to_model_type.get(nep_type )
    except FileNotFoundError:
        pass
        # logger.warning(f"Error: File {file_path} not found. Default model_type is 0")
    except Exception as e:
        logger.warning(f"An error occurred while parsing the file: {e}")

    return model_type

def get_xyz_nframe(path):
    if os.path.exists(path):
        with open(path, 'r',encoding="utf8") as file:
            nums = re.findall("^(\d+)$", file.read(), re.MULTILINE)
            return len(nums)
    return 0




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from PySide6.QtWidgets import QFrame, QGridLayout
from qfluentwidgets import BodyLabel, ToolTipFilter, ToolTipPosition, CheckBox

from NepTrainKit.core import CardManager, process_organic_clusters, get_clusters
from NepTrainKit.custom_widget import SpinBoxUnitInputFrame
from NepTrainKit.custom_widget.card_widget import MakeDataCard

@CardManager.register_card
class ShearMatrixCard(MakeDataCard):
    group = "Lattice"
    card_name = "Shear Matrix Strain"
    menu_icon = r":/images/src/images/scaling.svg"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Make Shear Matrix Strain")
        self.init_ui()

    def init_ui(self):
        self.setObjectName("shear_strain_card_widget")
        self.optional_frame = QFrame(self.setting_widget)
        self.optional_frame_layout = QGridLayout(self.optional_frame)
        self.optional_frame_layout.setContentsMargins(0, 0, 0, 0)
        self.optional_frame_layout.setSpacing(2)

        self.optional_label = BodyLabel("Optional", self.setting_widget)
        self.organic_checkbox = CheckBox("Identify organic", self.setting_widget)
        self.organic_checkbox.setChecked(False)
        self.symmetric_checkbox = CheckBox("Symmetric shear", self.setting_widget)
        self.symmetric_checkbox.setChecked(True)
        self.symmetric_checkbox.setToolTip("Apply shear symmetrically")
        self.symmetric_checkbox.installEventFilter(ToolTipFilter(self.symmetric_checkbox, 300, ToolTipPosition.TOP))
        self.optional_label.setToolTip("Treat organic molecules as rigid units")
        self.optional_label.installEventFilter(ToolTipFilter(self.optional_label, 300, ToolTipPosition.TOP))
        self.optional_frame_layout.addWidget(self.organic_checkbox, 0, 0, 1, 1)
        self.optional_frame_layout.addWidget(self.symmetric_checkbox, 1, 0, 1, 1)

        self.xy_label = BodyLabel("XY:", self.setting_widget)
        self.xy_frame = SpinBoxUnitInputFrame(self)
        self.xy_frame.set_input(["-", "% step:", "%"], 3, "float")
        self.xy_frame.setRange(-100, 100)
        self.xy_frame.set_input_value([-5, 5, 1])
        self.xy_label.setToolTip("XY shear strain range")
        self.xy_label.installEventFilter(ToolTipFilter(self.xy_label, 300, ToolTipPosition.TOP))

        self.yz_label = BodyLabel("YZ:", self.setting_widget)
        self.yz_frame = SpinBoxUnitInputFrame(self)
        self.yz_frame.set_input(["-", "% step:", "%"], 3, "float")
        self.yz_frame.setRange(-100, 100)
        self.yz_frame.set_input_value([-5, 5, 1])
        self.yz_label.setToolTip("YZ shear strain range")
        self.yz_label.installEventFilter(ToolTipFilter(self.yz_label, 300, ToolTipPosition.TOP))

        self.xz_label = BodyLabel("XZ:", self.setting_widget)
        self.xz_frame = SpinBoxUnitInputFrame(self)
        self.xz_frame.set_input(["-", "% step:", "%"], 3, "float")
        self.xz_frame.setRange(-100, 100)
        self.xz_frame.set_input_value([-5, 5, 1])
        self.xz_label.setToolTip("XZ shear strain range")
        self.xz_label.installEventFilter(ToolTipFilter(self.xz_label, 300, ToolTipPosition.TOP))

        self.settingLayout.addWidget(self.optional_label, 0, 0, 1, 1)
        self.settingLayout.addWidget(self.optional_frame, 0, 1, 1, 2)
        self.settingLayout.addWidget(self.xy_label, 1, 0, 1, 1)
        self.settingLayout.addWidget(self.xy_frame, 1, 1, 1, 2)
        self.settingLayout.addWidget(self.yz_label, 2, 0, 1, 1)
        self.settingLayout.addWidget(self.yz_frame, 2, 1, 1, 2)
        self.settingLayout.addWidget(self.xz_label, 3, 0, 1, 1)
        self.settingLayout.addWidget(self.xz_frame, 3, 1, 1, 2)

    def process_structure(self, structure):
        structure_list = []
        xy = self.xy_frame.get_input_value()
        yz = self.yz_frame.get_input_value()
        xz = self.xz_frame.get_input_value()
        identify_organic = self.organic_checkbox.isChecked()

        if identify_organic:
            clusters, is_organic_list = get_clusters(structure)

        xy_range = np.arange(xy[0], xy[1] + 0.001, xy[2])
        yz_range = np.arange(yz[0], yz[1] + 0.001, yz[2])
        xz_range = np.arange(xz[0], xz[1] + 0.001, xz[2])
        cell = structure.get_cell()
        symmetric = self.symmetric_checkbox.isChecked()

        for sxy in xy_range:
            for syz in yz_range:
                for sxz in xz_range:
                    new_structure = structure.copy()
                    shear_matrix = np.eye(3)
                    shear_matrix[0, 1] += sxy / 100
                    shear_matrix[1, 2] += syz / 100
                    shear_matrix[0, 2] += sxz / 100
                    if symmetric:
                        shear_matrix[1, 0] += sxy / 100
                        shear_matrix[2, 1] += syz / 100
                        shear_matrix[2, 0] += sxz / 100

                    new_cell = np.matmul(cell, shear_matrix)
                    new_structure.set_cell(new_cell, scale_atoms=True)
                    if identify_organic:
                        process_organic_clusters(structure, new_structure, clusters, is_organic_list)  # pyright:ignore
                    info_list = []
                    if abs(sxy) > 1e-8:
                        info_list.append(f"xy:{sxy}%")
                    if abs(syz) > 1e-8:
                        info_list.append(f"yz:{syz}%")
                    if abs(sxz) > 1e-8:
                        info_list.append(f"xz:{sxz}%")
                    info_str = "|".join(info_list)
                    new_structure.info["Config_type"] = new_structure.info.get("Config_type", "") + f" Shear({info_str},symmetric={symmetric})"
                    structure_list.append(new_structure)
        return structure_list

    def to_dict(self):
        data_dict = super().to_dict()
        data_dict["organic"] = self.organic_checkbox.isChecked()
        data_dict["symmetric"] = self.symmetric_checkbox.isChecked()
        data_dict["xy_range"] = self.xy_frame.get_input_value()
        data_dict["yz_range"] = self.yz_frame.get_input_value()
        data_dict["xz_range"] = self.xz_frame.get_input_value()
        return data_dict

    def from_dict(self, data_dict):
        super().from_dict(data_dict)
        self.organic_checkbox.setChecked(data_dict.get("organic", False))
        self.symmetric_checkbox.setChecked(data_dict.get("symmetric", True))
        self.xy_frame.set_input_value(data_dict.get("xy_range", [-5, 5, 1]))
        self.yz_frame.set_input_value(data_dict.get("yz_range", [-5, 5, 1]))
        self.xz_frame.set_input_value(data_dict.get("xz_range", [-5, 5, 1]))



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# Lazy exports for core.io to reduce import time and avoid cycles.
from __future__ import annotations

from typing import Any

__all__ = [
    # base
    'ResultData',
    # nep
    'NepTrainResultData', 'NepPolarizabilityResultData', 'NepDipoleResultData',
    # deepmd
    'DeepmdResultData', 'is_deepmd_path',
    # utils
    'get_nep_type',
    # registry helpers
    'load_result_data', 'register_result_loader', 'matches_result_loader',
]


def __getattr__(name: str) -> Any:  # PEP 562 lazy attribute loading
    if name == 'ResultData':
        from .base import ResultData as _T
        return _T
    if name in ('NepTrainResultData', 'NepPolarizabilityResultData', 'NepDipoleResultData'):
        from . import nep as _nep
        return getattr(_nep, name)
    if name in ('DeepmdResultData', 'is_deepmd_path'):
        from . import deepmd as _dp
        return getattr(_dp, name)
    if name == 'get_nep_type':
        from .utils import get_nep_type as _gt
        return _gt
    if name in ('load_result_data', 'register_result_loader', 'matches_result_loader'):
        from . import registry as _reg
        return getattr(_reg, name)
    raise AttributeError(name)



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/12/20 17:18
# @Author  : 兵
# @email    : 1747193328@qq.com
import json
import os.path

import numpy as np
from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QWidget, QGridLayout, QApplication
from ase import Atoms, Atom
from qfluentwidgets import HyperlinkLabel, BodyLabel, SubtitleLabel

from NepTrainKit.core import MessageManager, CardManager
from NepTrainKit.config import Config
from NepTrainKit.custom_widget import MakeWorkflowArea

from NepTrainKit.views import   ConsoleWidget


from NepTrainKit.version import __version__
from NepTrainKit import utils, get_user_config_path
from ase.io import read as ase_read



class MakeDataWidget(QWidget):
    """
微扰训练集制作
    """

    def __init__(self,parent=None):
        super().__init__(parent)
        self._parent = parent
        self.setObjectName("MakeDataWidget")
        self.setAcceptDrops(True)
        self.nep_result_data=None
        self.init_action()
        self.init_ui()
        self.dataset=None


    def dragEnterEvent(self, event):

        # 检查拖拽的内容是否包含文件
        if event.mimeData().hasUrls():
            event.acceptProposedAction()  # 接受拖拽事件

        else:
            event.ignore()  # 忽略其他类型的拖拽

    def dropEvent(self, event):
        # 获取拖拽的文件路径
        # print("dropEvent",event)
        urls = event.mimeData().urls()

        if urls:
            # 获取第一个文件路径
            structures_path = []
            for url in urls:
                file_path = url.toLocalFile()
                if (file_path.endswith(".xyz") or
                    file_path.endswith(".vasp") or
                    file_path.endswith(".cif")):
                    structures_path.append(file_path)

                elif file_path.endswith(".json"):
                    self.parse_card_config(file_path)
                else:
                    MessageManager.send_info_message("Only .xyz .vasp .cif or json files are supported for import.")
            if structures_path:
                self.load_base_structure(structures_path)

        # event.accept()

    def showEvent(self, event):
        if hasattr(self._parent,"load_menu"):
            self._parent.load_menu.addAction(self.load_card_config_action)  # pyright:ignore
        if hasattr(self._parent,"save_menu"):
            self._parent.save_menu.addAction(self.export_card_config_action)  # pyright:ignore

    def hideEvent(self, event):
        if hasattr(self._parent,"load_menu"):
            self._parent.load_menu.removeAction(self.load_card_config_action)  # pyright:ignore
        if hasattr(self._parent,"save_menu"):
            self._parent.save_menu.removeAction(self.export_card_config_action)   # pyright:ignore

    def init_action(self):
        self.export_card_config_action = QAction(QIcon(r":/images/src/images/save.svg"), "Export Card Config")
        self.export_card_config_action.triggered.connect(self.export_card_config)
        self.load_card_config_action = QAction(QIcon(r":/images/src/images/open.svg"), "Import Card Config")
        self.load_card_config_action.triggered.connect(self.load_card_config)
    def init_ui(self):

        self.gridLayout = QGridLayout(self)
        self.gridLayout.setObjectName("make_data_gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.workspace_card_widget = MakeWorkflowArea(self)
        self.setting_group=ConsoleWidget(self)
        self.setting_group.runSignal.connect(self.run_card)
        self.setting_group.stopSignal.connect(self.stop_run_card)
        self.setting_group.newCardSignal.connect(self.add_card)

        self.path_label = HyperlinkLabel(self)
        self.path_label.setFixedHeight(30)  # 设置状态栏的高度
        user_config_path = get_user_config_path()
        self.path_label.setText("Folder for Custom Cards  ")

        self.path_label.setUrl(f"file:///{user_config_path}/cards")

        self.dataset_info_label = BodyLabel(self)
        self.dataset_info_label.setFixedHeight(30)  # 设置状态栏的高度

        self.gridLayout.addWidget(self.setting_group, 0, 0, 1, 2)
        self.gridLayout.addWidget(self.workspace_card_widget, 1, 0, 1, 2)
        self.gridLayout.addWidget(self.dataset_info_label, 2, 0, 1, 1)
        self.gridLayout.addWidget(self.path_label, 2, 1, 1, 1,alignment=Qt.AlignmentFlag.AlignRight)
        self.setLayout(self.gridLayout)

    def load_base_structure(self,paths):

        structures_list = []
        for path  in paths:
            atoms  = ase_read(path,":")
            #ase有时候会将字符串解析成数组或者int  这里转换成str
            for atom in atoms:
                if isinstance(atom, Atom):
                    continue

                if 'config_type' in atom.info:
                    atom.info["Config_type"]=atom.info["config_type"]
                    del atom.info["config_type"]



                if isinstance(atom.info.get("Config_type"),np.ndarray):
                    if atom.info["Config_type"].size==0:

                        atom.info["Config_type"] = Config.get("widget", "default_config_type", "neptrainkit")
                    else:
                        atom.info["Config_type"]=" ".join(atom.info["Config_type"])

                else:
                    atom.info["Config_type"]=str(atom.info.get("Config_type", Config.get("widget", "default_config_type", "neptrainkit")))

                structures_list.append(atom)

        self.dataset=structures_list
        MessageManager.send_success_message(f"success load {len(structures_list)} structures.")
        self.dataset_info_label.setText(f" Success load {len(structures_list)} structures.")
    def open_file(self):
        path = utils.call_path_dialog(self,"Please choose the structure files",
                                      "selects",file_filter="Structure Files (*.xyz *.vasp *.cif)")

        if path:
            self.load_base_structure(path)
    def _export_file(self,path):
        if os.path.exists(path):
            os.remove(path)
        with open(path, "w",encoding="utf8") as file:
            for card in self.workspace_card_widget.cards:
                if card.check_state:

                    card.write_result_dataset(file,append=True)


    def export_file(self):


        path = utils.call_path_dialog(self, "Choose a file save location", "file",default_filename="make_dataset.xyz")
        if path:
            thread = utils.LoadingThread(self, show_tip=True, title="Exporting data")
            thread.start_work(self._export_file, path)
    def run_card(self):
        if not  self.dataset  :
            MessageManager.send_info_message("Please import the structure file first. You can drag it in directly or import it from the upper left corner!")
            return
        self.stop_run_card()
        first_card=self._next_card(-1)
        if first_card:
            first_card.dataset = self.dataset

            first_card.runFinishedSignal.connect(self._run_next_card)
            first_card.run()
        else:
            MessageManager.send_info_message("No card selected. Please select a card in the workspace.")
    def _next_card(self,current_card_index=-1):

        cards=self.workspace_card_widget.cards
        if current_card_index+1 >=len(cards):
            return None
        current_card_index+=1
        for i,card in enumerate(cards[current_card_index:]):

            if card.check_state:
                card.index=i+current_card_index
                return card
            else:
                continue
        return None
    def _run_next_card(self,current_card_index):

        cards=self.workspace_card_widget.cards
        current_card=cards[current_card_index]
        current_card.runFinishedSignal.disconnect(self._run_next_card)

        next_card=self._next_card(current_card_index )
        if current_card.result_dataset and next_card:
            next_card.set_dataset(current_card.result_dataset)

            next_card.runFinishedSignal.connect(self._run_next_card)
            next_card.run()
        else:
            MessageManager.send_success_message("Perturbation training set created successfully.")
    def stop_run_card(self):
        for card in self.workspace_card_widget.cards:
            try:
                card.runFinishedSignal.disconnect(self._run_next_card)
            except:
                pass
            card.stop()

    def add_card(self,card_name):


        if card_name not in CardManager.card_info_dict:
            MessageManager.send_warning_message("no card")
            return None
        card=CardManager.card_info_dict[card_name](self)
        self.workspace_card_widget.add_card(card)
        return card

    def export_card_config(self):
        cards=self.workspace_card_widget.cards
        if not cards:
            MessageManager.send_warning_message("No cards in workspace.")

            return

        path = utils.call_path_dialog(self, "Choose a file save location", "file", default_filename="card_config.json")
        if path:
            config={}
            config["software_version"]=__version__
            config["cards"]=[]
            for card in cards:
                config["cards"].append(card.to_dict())


            with open(path, "w",encoding="utf-8") as file:
                json.dump(config, file, indent=4,ensure_ascii=False)
            MessageManager.send_success_message("Card configuration exported successfully.")
    def load_card_config(self):
        path = utils.call_path_dialog(self, "Choose a card configuration file", "select" )
        if path:

            self.parse_card_config(path)
    def parse_card_config(self,path):
            try:
                with open(path, "r",encoding="utf-8") as file:
                    config = json.load(file)
            except:
                MessageManager.send_warning_message("Invalid card configuration file.")
                return
            self.workspace_card_widget.clear_cards()
            cards=config.get("cards")
            for card in cards:
                name=card.get("class")
                card_widget=self.add_card(name)
                if card_widget is not None:
                    card_widget.from_dict(card)


if __name__ == "__main__":
    import sys
    app = QApplication(sys.argv)
    from NepTrainKit.core import Config
    Config()

    window = MakeDataWidget()
    window.resize( 800,600)
    window.show()
    sys.exit(app.exec())



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/21 14:45
# @Author  : 兵
# @email    : 1747193328@qq.com
from __future__ import annotations
import glob
import json
import os
import re
from copy import deepcopy
from pathlib import Path
from functools import cached_property
from typing import Any,IO

import numpy as np
import numpy.typing as npt
from ase import neighborlist
from ase import Atoms as _ASEAtoms
from ase.geometry import find_mic
from loguru import logger
from scipy.sparse.csgraph import connected_components
from collections import defaultdict, Counter
from NepTrainKit import utils, module_path
from NepTrainKit.config import Config



with open(os.path.join(module_path, "Config/ptable.json"), "r", encoding="utf-8") as f:
    table_info = json.loads(f.read())


atomic_numbers={elem_info["symbol"]:elem_info["number"] for elem_info in table_info.values()}

class Structure:
    """
    extxyz格式的结构类
    原子坐标是笛卡尔坐标
    """
    def __init__(self,
                 lattice: list[float]|npt.NDArray[np.float32],
                 atomic_properties:dict[str,npt.NDArray[np.float32]],
                 properties:list[dict[str,str]],
                 additional_fields:dict[str,Any]  ):
        """

        :param lattice: 晶格矩阵
        :param atomic_properties:
            和原子对应的属性 比如{"pos":np.array(),"forces":np.array(),"species":np.array()}
        :param properties: [{'name': "pos", 'type': 'R', 'count': 3},...]
        :param additional_fields: xyz的第二行的信息 比如{"energy":3}
        """
        super().__init__()
        self.properties = properties
        self.lattice = np.array(lattice,dtype=np.float32).reshape((3,3))  # Optional: Lattice vectors
        self.atomic_properties = atomic_properties
        self.additional_fields = additional_fields
        if "Config_type" not in self.additional_fields.keys():
            self.additional_fields["Config_type"] = Config.get("widget", "default_config_type", "neptrainkit")
        if "force" in self.atomic_properties.keys():
            self.force_label="force"
        else:
            self.force_label = "forces"
        #预加载下属性
        self.formula
    @property
    def tag(self)->str:
        """Alias for the ``Config_type`` additional field."""
        return self.additional_fields.get("Config_type", "")

    @tag.setter
    def tag(self, value):
        self.additional_fields["Config_type"] = value
    def get_prop_key(self,additional_fields=True,atomic_properties=True)->list[str]:
        """
        获取整个结构中，所有的属性key值
        :param additional_fields:
        :param atomic_properties:
        :return: ["pos","energy",...]
        """
        keys=[]
        if additional_fields:
            keys.extend(self.additional_fields.keys())
        if atomic_properties:
            keys.extend(self.atomic_properties.keys())
        return keys
    def remove_atomic_properties(self,key:str):
        if key in self.atomic_properties:
            self.atomic_properties.pop(key)
            for prop in self.properties:
                if prop["name"]==key:
                    self.properties.remove(prop)
                    break

    def __len__(self):
        return len(self.elements)

    @classmethod
    def read_xyz(cls, filename:str) -> Structure:
        with open(filename, 'r',encoding="utf8") as f:
            structure = cls.parse_xyz(f.read())
        return structure

    @staticmethod
    def iter_read_multiple(filename: str, cancel_event=None):
        """
        Incrementally read a multi-structure XYZ file and yield Structure objects.
        If cancel_event is provided and set(), stop early.
        """
        with open(filename, "r",encoding="utf8") as file:
            while True:
                if cancel_event is not None and getattr(cancel_event, "is_set", None) and cancel_event.is_set():
                    return
                num_atoms_line = file.readline()
                if not num_atoms_line:
                    break
                num_atoms_line = num_atoms_line.strip()
                if not num_atoms_line:
                    continue

                num_atoms = int(num_atoms_line)

                global_line = file.readline()
                if not global_line:
                    break
                structure_lines = [num_atoms_line, global_line.rstrip()]
                for _ in range(num_atoms):
                    if cancel_event is not None and getattr(cancel_event, "is_set", None) and cancel_event.is_set():
                        return
                    line = file.readline()
                    if not line:
                        break
                    structure_lines.append(line.rstrip())

                yield Structure.parse_xyz(structure_lines)




    @property
    def cell(self):
        return self.lattice

    @property
    def volume(self):
        return np.abs(np.linalg.det(self.lattice))

    @property
    def abc(self):
        """Return lattice vector lengths (a, b, c)."""
        return np.linalg.norm(self.lattice, axis=1)

    @property
    def angles(self):
        """Return lattice angles (alpha, beta, gamma) in degrees."""
        a_vec, b_vec, c_vec = self.lattice

        def _angle(v1, v2):
            cos_ang = np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))
            cos_ang = np.clip(cos_ang, -1.0, 1.0)
            return np.degrees(np.arccos(cos_ang))

        alpha = _angle(b_vec, c_vec)
        beta = _angle(a_vec, c_vec)
        gamma = _angle(a_vec, b_vec)
        return np.array([alpha, beta, gamma], dtype=np.float32)

    @property
    def numbers(self):
        return [atomic_numbers[element] for element in self.elements ]
    @property
    def spin_num(self)->int:
        if  "force_mag" not in self.atomic_properties :
            return 0
        mag=self.atomic_properties["force_mag"]
        count = np.sum(~np.all(mag == 0, axis=1))
        return count
    @cached_property
    def formula(self):
        #这种形式会导致化学式过长 比如有机分子环境下
        # diffs = np.diff(self.numbers)
        # # 找到变化的地方，包含第一个元素的起始位置
        # change_points = np.where(diffs != 0)[0] + 1
        # # 在变化的位置之间进行分段
        # segments = np.split(self.elements, change_points)
        # # 格式化每个段落
        # result = [f"{segment[0]}{len(segment)}" for segment in segments]
        # return "".join(result)
        # 改成下面的形式
        # formula = ""
        # elems={}
        # for element in self.elements:
        #     if element in elems.keys():
        #         elems[element]+=1
        #     else:
        #         elems[element]=1
        # for element,count in elems.items():
        #     formula+=element+str(count)
        # return formula
        return self.__get_formula(sub=False)



    @cached_property
    def html_formula(self)->str:

        return self.__get_formula(sub=True)

    def __get_formula(self, sub=False)->str:
        # priority = {'C': 0, 'H': 1}

        # 优先级：C → H → 按原子序数 → 未知元素按字母
        def priority(sym):
            if sym == 'C':
                return (0, 0)
            if sym == 'H':
                return (0, 1)
            z = atomic_numbers.get(sym, 999)
            return (1, z) if z != 999 else (2, sym)

        cnt = Counter(self.elements)
        count2 = dict(sorted(cnt.items(), key=lambda kv: priority(kv[0])))
        if sub:
            formula = ''.join(symb + ("<sub>" + str(n) + "</sub>" if n > 1 else '')
                              for symb, n in count2.items())
        else:

            formula = ''.join(symb + (str(n) if n > 1 else '')
                              for symb, n in count2.items())
        return formula
    @property
    def per_atom_energy(self):
        return self.energy/self.num_atoms
    @property
    def energy(self):
        return self.additional_fields["energy"]
    @energy.setter
    def energy(self,new_energy:float):
        self.additional_fields["energy"] = new_energy
    @property
    def forces(self):
        return self.atomic_properties[self.force_label]
    @forces.setter
    def forces(self,arr:npt.NDArray[np.float32]):
        has_forces=[i["name"]==self.force_label for i in self.properties]
        if not any(has_forces):
            self.properties.append({'name': self.force_label, 'type': 'R', 'count': 3})

        self.atomic_properties[self.force_label] = arr

    @property
    def virial(self):
        try:
            vir =self.additional_fields["virial"]
        except:
            # 检查下有没有压强
            try:
                vir = self.additional_fields["stress"]  * self.volume * -1
            except:
                raise ValueError("No virial or stress data")
        return vir
    @virial.setter
    def virial(self,new_virial:npt.NDArray[np.float32]):
        self.additional_fields["virial"] = new_virial

    @property
    def nep_virial(self):

        vir=self.virial
        return vir[[0,4,8,1,5,6]]/self.num_atoms

    @property
    def nep_dipole(self):
        dipole=np.array(self.dipole.split(" "),dtype=np.float32)
        return dipole/self.num_atoms

    @property
    def nep_polarizability(self):
        vir = np.array(self.pol.split(" "), dtype=np.float32)
        return vir[[0,4,8,1,5,6]] / self.num_atoms

    def get_chemical_symbols(self):
        return self.elements

    @property
    def elements(self):
        return self.atomic_properties['species']

    @property
    def positions(self):
        return self.atomic_properties['pos']

    @property
    def num_atoms(self):
        return len(self.elements)

    def copy(self):
        return deepcopy(self)

    def set_lattice(self, new_lattice: npt.NDArray[np.float32],in_place=False):
        """
        根据新晶格缩放原子位置，支持原地修改或返回新对象。

        :param new_lattice: 新晶格矩阵（3x3 numpy 数组）
        :param in_place: 是否修改当前对象（默认 False，返回新对象）
        :return: 更新后的 Structure 对象（若 in_place=True，则返回 self）
        """
        target = self if in_place else self.copy()
        old_lattice = target.lattice
        old_positions = target.positions

        # 计算变换矩阵（参考 ASE）
        M = np.linalg.solve(old_lattice, new_lattice)
        new_positions = old_positions @ M

        # 更新晶格和坐标
        target.lattice = new_lattice
        target.atomic_properties['pos'] = new_positions

        return target

    def supercell(self, scale_factor, order="atom-major", tol=1e-5)->Structure:
        """
        按指定比例因子扩展晶胞，参考 ASE 的高效实现。
        保持晶格角度不变，并支持按元素排序。

        :param scale_factor: 扩展比例因子（标量或长度为3的数组，对应 a、b、c 方向）
        :param order: 原子排序方式，"cell-major"（默认）或 "atom-major"
        :param tol: 数值容差，用于边界检查
        :return: 扩展后的新 Structure 对象
        :raises ValueError: 如果 scale_factor 无效
        """
        # 输入验证和标准化
        scale_factor = np.asarray(scale_factor, dtype=np.float32)
        if scale_factor.size == 1:
            scale_factor = np.full(3, scale_factor)
        if scale_factor.size != 3:
            raise ValueError("scale_factor 必须是标量或长度为3的数组")
        if scale_factor.min() < 1:
            raise ValueError("scale_factor 必须大于等于 1")

        # 输入验证（同上）
        scale_factor = np.asarray(scale_factor, dtype=np.int64)  # 限制为整数扩展

        # 计算新晶格（各方向独立扩展）
        new_lattice = self.lattice * scale_factor[:, None]

        # 转换原始坐标到分数坐标并包裹
        inv_orig_lattice = np.linalg.inv(self.lattice)
        frac_pos = self.positions @ inv_orig_lattice
        frac_pos = frac_pos % 1.0  # 严格包裹

        # 生成扩展网格（明确轴向顺序）
        n_a, n_b, n_c = scale_factor
        # 生成平移向量时确保a方向优先
        offsets_a = np.arange(n_a)[:, None] * np.array([1, 0, 0])  # a方向偏移
        offsets_b = np.arange(n_b)[:, None] * np.array([0, 1, 0])  # b方向偏移
        offsets_c = np.arange(n_c)[:, None] * np.array([0, 0, 1])  # c方向偏移

        # 计算所有偏移组合（a方向最外层循环）
        full_offsets = (offsets_a[:, None, None] +
                        offsets_b[None, :, None] +
                        offsets_c[None, None, :]).reshape(-1, 3)

        # 扩展分数坐标
        expanded_frac = frac_pos[:, None, :] + full_offsets[None, :, :]
        expanded_frac = expanded_frac.reshape(-1, 3) / scale_factor  # 归一化到新分数坐标

        # 转换到新笛卡尔坐标
        new_positions = expanded_frac @ new_lattice

        # 元素扩展（保持a方向优先顺序）
        if order == "cell-major":
            new_elements = np.tile(self.elements, int(np.prod(scale_factor)))
        elif order == "atom-major":
            new_elements = np.repeat(self.elements, np.prod(scale_factor))
        else:
            raise ValueError( )

        # 更新结构信息
        atomic_properties = {}
        atomic_properties['pos'] = new_positions.astype(np.float32)
        atomic_properties['species'] = new_elements

        properties=[{'name': 'species', 'type': 'S', 'count': 1}, {'name': 'pos', 'type': 'R', 'count': 3}]
        # 设置周期性边界条件（假设与原始一致）
        additional_fields={}
        additional_fields['pbc'] = self.additional_fields.get('pbc', "T T T")
        additional_fields["Config_type"] =self.additional_fields.get('Config_type', "")+f" super cell({scale_factor})"

        return Structure(new_lattice, atomic_properties, properties, additional_fields)
    def adjust_reasonable(self, coefficient=0.7)->bool:
        """
        根据传入系数 对比共价半径和实际键长，
        如果实际键长小于coefficient*共价半径之和，判定为不合理结构 返回False
        否则返回 True
        :param coefficient: 系数
        :return:

        """
        distance_info = self.get_mini_distance_info()
        for elems, bond_length in distance_info.items():
            elem0_info = table_info[str(atomic_numbers[elems[0]])]
            elem1_info = table_info[str(atomic_numbers[elems[1]])]

            # 相邻原子距离小于共价半径之和×系数就选中
            if (elem0_info["radii"] + elem1_info["radii"]) * coefficient > bond_length * 100:
                return False
        return True





    # 在序列化时使用 __getstate__ 进行处理
    def __getstate__(self):
        # 返回对象的状态字典，这里可以控制哪些属性需要序列化

        state = self.__dict__.copy()

        return state

    # 反序列化时使用 __setstate__
    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getattr__(self, item):

        if item in self.additional_fields.keys():
            return self.additional_fields[item]
        elif item in self.atomic_properties.keys():
            return self.atomic_properties[item]
        else:
            raise AttributeError


    @classmethod
    def parse_xyz(cls, lines:list[str]|str)->Structure:
        """
        Parse a single structure from a list of lines.
        """
        if isinstance(lines, str):
            lines = lines.strip().split('\n')
        # Parse the second line (global properties)
        global_properties = lines[1].strip()
        lattice, properties, additional_fields = cls._parse_global_properties(global_properties)
        array = np.array([line.split() for line in lines[2:]],dtype=object )

        atomic_properties = {}
        index = 0

        for prop in properties:

            _info = array[:, index:index + prop["count"]]
            #
            # _info =[row[index:index + prop["count"]] for row in array]

            if prop["type"] == "S":
                pass
                _info=_info.astype( np.str_)

            elif prop["type"] == "R":
                _info=_info.astype( np.float32)

            elif prop["type"] == "L":
                _info=_info.astype( np.bool_)
            elif  prop["type"] == "I":
                _info=_info.astype( np.int8)

            else:
                pass
            if prop["count"] == 1:
                _info = _info.flatten()
            else:
                _info = _info.reshape((-1, prop["count"]))

            atomic_properties[prop["name"]] = _info
            index += prop["count"]
        del array

        # return
        return cls(lattice, atomic_properties, properties, additional_fields)

    @classmethod
    def _parse_global_properties(cls, line:str)->tuple[list[float],list[dict[str,Any]],dict[str,Any]]:
        """
        Parse global properties from the second line of an XYZ block.
        """
        pattern = r'(\w+)=\s*"([^"]+)"|(\w+)=([\S]+)'
        matches = re.findall(pattern, line)
        properties = []
        lattice = None
        additional_fields = {}

        for match in matches:
            key = match[0] or match[2]
            # key=key.capitalize()
            value = match[1] or match[3]

            if key.capitalize()  == "Lattice":
                lattice = list(map(float, value.split()))
            elif key.capitalize()  == "Properties":
                # Parse Properties details
                properties = cls._parse_properties(value)
            else:

                if '"' in value:

                    value = value.strip('"')  # 去掉引号
                else:
                    try:
                        value = float(value)
                    except Exception as e:
                        value = value
                if key == "config_type" or key == "Config_type":
                    # 这里是为了后面的Config搜索做统一
                    key = "Config_type"
                    value=str(value)
                if key.lower() in ("energy", "pbc","virial","stress"):
                    key=key.lower()
                if key =="virial" or key =="stress":
                    value= np.array(str(value).split(" "), dtype=np.float32)   # pyright:ignore
                additional_fields[key] = value
                # print(additional_fields)
        return lattice, properties, additional_fields

    @staticmethod
    def _parse_properties(properties_str)->list[dict[str,Any]]:
        """
        Parse `Properties` attribute string to extract atom-specific fields.
        """
        tokens = properties_str.split(":")
        parsed_properties = []
        i = 0
        while i < len(tokens):
            name = tokens[i]
            dtype = tokens[i + 1]
            count = int(tokens[i + 2]) if i + 2 < len(tokens) else 1
            parsed_properties.append({"name": name, "type": dtype, "count": count})
            i += 3
        return parsed_properties

    @staticmethod
    @utils.timeit
    def read_multiple(filename:str ):
        """
        Read a multi-structure XYZ file and return a list of Structure objects.
        """


        # data_to_process = []
        structures = []

        with open(filename, "r",encoding="utf8") as file:
            while True:
                num_atoms_line = file.readline()
                if not num_atoms_line:
                    break
                num_atoms_line = num_atoms_line.strip()
                if not num_atoms_line:
                    continue
                num_atoms = int(num_atoms_line)
                structure_lines = [num_atoms_line, file.readline().rstrip()]  # global properties
                for _ in range(num_atoms):
                    line = file.readline()
                    if not line:
                        break
                    structure_lines.append(line.rstrip())

                structure = Structure.parse_xyz(structure_lines)
                structures.append(structure)
                del structure_lines

        return structures

    def write(self, file:IO):
        """
        Write the current structure to an XYZ file.
        """

        # Write number of atoms
        file.write(f"{self.num_atoms}\n")

        # Write global properties
        global_line = []
        if self.lattice.size!=0:
            global_line.append(f'Lattice="' + ' '.join(f"{x}" for x in self.cell.flatten()) + '"')

        props = ":".join(f"{p['name']}:{p['type']}:{p['count']}" for p in self.properties)
        global_line.append(f"Properties={props}")
        for key, value in self.additional_fields.items():

            if isinstance(value, (float, int,np.number)):
                global_line.append(f"{key}={value}")
            elif isinstance(value, np.ndarray):
                value_str = " ".join(map(str, value.flatten()))
                global_line.append(f'{key}="{value_str}"')
            else:
                global_line.append(f'{key}="{value}"')
        file.write(" ".join(global_line) + "\n")

        for row in range(self.num_atoms):
            line = ""
            for prop  in self.properties :
                if prop["count"] == 1:
                    values=[self.atomic_properties[prop["name"]][row]]
                else:
                    values= self.atomic_properties[prop["name"]][row, :]

                if prop["type"] == 'S':  # 字符串类型
                    line += " ".join([f"{x }" for x in values]) + " "

                elif prop["type"] == 'R':  # 浮点数类型
                    line += " ".join([f"{x:.10g}" for x in values]) + " "
                else:
                    line += " ".join([f"{x}" for x in values]) + " "
            file.write(line.strip() + "\n")

    def get_all_distances(self):
        return  calculate_pairwise_distances(self.cell, self.positions, False)

    def get_mini_distance_info(self):
        """
        返回原子对之间的最小距离
        """
        # 使用分块 + 最小镜像向量，避免 NxN 全矩阵占用
        cell = np.asarray(self.cell, dtype=np.float32).reshape(3, 3)
        inv_cell = np.linalg.inv(cell)
        coords = np.asarray(self.positions, dtype=np.float32).reshape(-1, 3)
        frac = coords @ inv_cell  # 转为分数坐标

        symbols = np.asarray([str(s) for s in self.elements])
        # 对元素做编码，便于聚合
        uniq_elems, codes = np.unique(symbols, return_inverse=True)
        E = uniq_elems.shape[0]
        # 每个元素的原子下标列表
        indices_by_code: list[np.ndarray] = [np.where(codes == c)[0] for c in range(E)]

        # 上三角最小距离矩阵（元素对）
        min_mat = np.full((E, E), np.inf, dtype=np.float32)

        # 预先计算度量张量，避免构造笛卡尔位移
        metric = cell @ cell.T  # 3x3

        N = frac.shape[0]
        # 自适应分块，尽量控制内存（~80MB 级别）
        # 估算单块开销 ~ (bs*N*3 + bs*N) * 4 bytes
        if N == 0:
            return {}
        target_bytes = 80 * 1024 * 1024
        per_pair_bytes = 4.0 * (3.0 + 1.0)  # df(3) + d(1)
        block_size = int(max(128, min(1024, target_bytes / (per_pair_bytes * N))))

        for i0 in range(0, N, block_size):
            i1 = min(N, i0 + block_size)
            df = frac[i0:i1, None, :] - frac[None, :, :]  # (bs, N, 3)
            # 最小镜像 [-0.5,0.5)
            df -= np.round(df)
            # d^2 = df @ metric @ df^T （逐元素）
            # (bs,N,3) x (3,3) -> (bs,N,3) ，再与 df 做 Einstein 求和
            tmp = df @ metric  # (bs,N,3)
            d2 = np.einsum('ijk,ijk->ij', df, tmp, dtype=np.float32)
            # 清除对角线（自身距离）
            for row, ii in enumerate(range(i0, i1)):
                d2[row, ii] = np.inf
            # 对每一行（固定 i）聚合到对应元素对的最小值
            for row, ii in enumerate(range(i0, i1)):
                ci = int(codes[ii])
                drow = d2[row]
                # 针对每个元素代码聚合最小值
                for cj in range(E):
                    idxs = indices_by_code[cj]
                    if idxs.size == 0:
                        continue
                    # 最小距离
                    m = float(np.min(drow[idxs]))
                    if not np.isfinite(m):
                        continue
                    # 归入有序的 (min,max) 上三角
                    a, b = (ci, cj) if ci <= cj else (cj, ci)
                    if m < min_mat[a, b]:
                        min_mat[a, b] = m

        # 组装结果：以元素对字符串元组为键，最小距离为值
        out: dict[tuple[str, str], float] = {}
        for a in range(E):
            for b in range(a, E):
                val = float(min_mat[a, b])
                if np.isfinite(val):
                    out[(str(uniq_elems[a]), str(uniq_elems[b]))] = np.sqrt(val, dtype=np.float32).item()
        return out

    def get_bond_pairs(self):
        """
        返回在范围内的所有键长
        """
        i, j = np.triu_indices(len(self), k=1)
        pos = np.array(self.positions)
        diff = pos[i] - pos[j]
        upper_distances = np.linalg.norm(diff, axis=1)
        covalent_radii = np.array([table_info[str(n)]["radii"] / 100 for n in self.numbers])
        radius_sum = covalent_radii[i] + covalent_radii[j]
        bond_mask = (upper_distances < radius_sum * 1.15)
        bond_pairs = [(i[k], j[k]) for k in np.where(bond_mask)[0]]
        return bond_pairs

    def get_bad_bond_pairs(self, coefficient=0.8):
        """
        根据键长阈值判断
        返回所有的非物理键长
        """
        i, j = np.triu_indices(len(self), k=1)
        distances = self.get_all_distances()
        upper_distances = distances[i, j]
        covalent_radii = np.array([table_info[str(n)]["radii"] / 100 for n in self.numbers])
        radius_sum = covalent_radii[i] + covalent_radii[j]
        bond_mask = (upper_distances < radius_sum * coefficient)
        bad_bond_pairs = [(i[k], j[k]) for k in np.where(bond_mask)[0]]
        return bad_bond_pairs



def calculate_pairwise_distances(lattice_params:npt.NDArray[np.float32],
                                 atom_coords:npt.NDArray[np.float32],
                                 fractional=True,
                                 block_size:int=2048
                                 )->npt.NDArray[np.float32]:
    """
    计算晶体中所有原子对之间的距离，考虑周期性边界条件

    参数:
    lattice_params: 晶格参数，3x3 numpy array 表示晶格向量 (a, b, c)
    atom_coords: 原子坐标，Nx3 numpy array
    fractional: 是否为分数坐标 (True) 或笛卡尔坐标 (False)

    返回:
    distances: NxN numpy array，所有原子对之间的距离
    """


    # Convert to fractional coordinates for minimum image without enumerating 27 images
    cell = np.asarray(lattice_params, dtype=np.float32).reshape(3, 3)
    coords = np.asarray(atom_coords, dtype=np.float32).reshape(-1, 3)
    if fractional:
        frac = coords.astype(np.float32)
    else:
        # frac = cart @ inv(cell)
        inv_cell = np.linalg.inv(cell)
        frac = coords @ inv_cell

    N = frac.shape[0]
    dmat = np.empty((N, N), dtype=np.float32)
    # Process in row blocks to reduce peak memory from O(N^2*27)
    for i0 in range(0, N, max(1, int(block_size))):
        i1 = min(N, i0 + int(block_size))
        # fractional differences with MIC wrapping into [-0.5, 0.5)
        df = frac[i0:i1, None, :] - frac[None, :, :]
        df -= np.round(df)
        # back to Cartesian and compute norms
        dr = df @ cell
        d = np.sqrt(np.sum(dr * dr, axis=2), dtype=np.float32)
        dmat[i0:i1, :] = d
    np.fill_diagonal(dmat, 0.0)
    return dmat



# 判断团簇是否为有机分子
def is_organic_cluster(symbols:list[str]) -> bool:
    has_carbon = 'C' in symbols
    organic_elements = {'H', 'O', 'N', 'S', 'P'}
    has_organic_elements = any(symbol in organic_elements for symbol in symbols)
    return has_carbon and has_organic_elements

# # 识别结构中的团簇


def get_clusters(structure):
    """

    :param structure:  ase.Atoms
    :return:
    """
    cutoff = neighborlist.natural_cutoffs(structure)
    nl = neighborlist.NeighborList(cutoff, self_interaction=False, bothways=True)
    nl.update(structure)
    matrix = nl.get_connectivity_matrix()
    n_components, component_list = connected_components(matrix)

    component_array = np.array(component_list)
    all_symbols = [atom.symbol for atom in structure]

    clusters = []
    is_organic_list = []
    for i in range(n_components):
        cluster_indices = np.where(component_array == i)[0].tolist()
        cluster_symbols = [all_symbols[j] for j in cluster_indices]
        clusters.append(cluster_indices)
        is_organic_list.append(is_organic_cluster(cluster_symbols))

    return clusters, is_organic_list
# 解包跨越边界的分子


def unwrap_molecule(structure, cluster_indices):
    pos = structure.positions[cluster_indices]
    cell = structure.cell
    ref_pos = pos[0]

    # 所有位置相对 ref_pos 的位移
    delta = pos - ref_pos

    # 求解最小镜像向量
    inv_cell = np.linalg.inv(cell.T)  # 提前计算逆矩阵
    frac_delta = np.dot(delta, inv_cell)
    frac_delta -= np.round(frac_delta)
    mic_delta = np.dot(frac_delta, cell.T)

    unwrapped_pos = ref_pos + mic_delta
    return unwrapped_pos


# 封装循环部分：处理有机分子团簇
def process_organic_clusters(structure, new_structure, clusters, is_organic_list):
    """处理有机分子团簇并更新原子位置"""

    for cluster_indices, is_organic in zip(clusters, is_organic_list):
        if is_organic:
            # 解包分子
            unwrapped_pos = unwrap_molecule(structure, cluster_indices)

            # 计算解包后质心
            center_unwrapped = np.mean(unwrapped_pos, axis=0)

            # 将质心转换到分数坐标并映射回晶胞内
            scaled_center = np.dot(center_unwrapped, np.linalg.inv(structure.cell)) % 1.0
            center_original = np.dot(scaled_center, structure.cell)

            # 计算原子相对于质心的位移
            delta_pos = unwrapped_pos - center_unwrapped

            # 在新晶胞中计算新质心
            center_new = np.dot(scaled_center, new_structure.cell)

            # 新位置 = 新质心 + 原始位移
            pos_new = center_new + delta_pos

            # 更新原子位置
            new_structure.positions[cluster_indices] = pos_new
    new_structure.wrap()



def _load_npy_structure(folder, cancel_event=None):
    structures = []
    type_map_path = os.path.join(folder, "type_map.raw")
    type_path = os.path.join(folder, "type.raw")

    if  not os.path.exists(type_path):
        return structures


    # Load once and reuse
    type_ = np.loadtxt(type_path, dtype=int,ndmin=1)

    if   os.path.exists(type_map_path) :
        type_map = np.loadtxt(type_map_path, dtype=str, ndmin=1)
    else:
        type_map=np.array([f"Type_{i+1}" for i in range(np.max(type_)+1)], dtype=str, ndmin=1)
    # Use np.array and list comprehension for faster element mapping

    # print(type_path,type_map,type_)
    elem_list = type_map[type_]

    atoms_num = len(elem_list)
    nopbc = os.path.isfile(os.path.join(folder, "nopbc"))

    sets = sorted(glob.glob(os.path.join(folder, "set.*")))
    dataset_dict = {}

    # Load data from files and use np.concatenate instead of vstack
    for _set in sets:
        if cancel_event is not None and getattr(cancel_event, "is_set", None) and cancel_event.is_set():
            return structures
        for data_path in Path(_set).iterdir():
            if cancel_event is not None and getattr(cancel_event, "is_set", None) and cancel_event.is_set():
                return structures
            key = data_path.stem
            data = np.load(data_path)
            if data.ndim == 1:
                data=data.reshape(data.shape[0], -1)
            if key in dataset_dict:
                dataset_dict[key].append(data)  # Collect in lists for later concatenation
            else:
                dataset_dict[key] = [data]

    config_type = os.path.basename(folder)

    # Efficient concatenation outside the loop
    for key in dataset_dict.keys():
        dataset_dict[key] = np.concatenate(dataset_dict[key], axis=0)
    logger.debug(f"load {dataset_dict['box'].shape[0]} structures from {folder}" )
    for index in range(dataset_dict["box"].shape[0]):
        if cancel_event is not None and getattr(cancel_event, "is_set", None) and cancel_event.is_set():
            break
        box = dataset_dict["box"][index].reshape(3, 3)
        coords = dataset_dict["coord"][index].reshape(-1, 3)

        properties = [
            {"name": "species", "type": "S", "count": 1},
            {"name": "pos", "type": "R", "count": 3},
        ]
        info = {
            "species": elem_list,
            "pos": coords,
        }
        additional_fields = {"Config_type": config_type}
        additional_fields["pbc"] = "F F F" if nopbc else "T T T"

        # Optimize adding properties and additional fields
        for key in dataset_dict.keys():
            if key not in ["box", "coord"]:
                prop = dataset_dict[key][index]
                count = prop.shape[0]

                if count > atoms_num and key!="virial":
                    col = count // atoms_num

                    info[key] = prop.reshape((-1, col))
                    properties.append({"name": key, "type": "R", "count": col})
                else:
                    if count == 1:
                        additional_fields[key] = prop[0]
                    else:

                        additional_fields[key] = prop.flatten()

        structure = Structure(lattice=box, atomic_properties=info, properties=properties,
                              additional_fields=additional_fields)
        structures.append(structure)

    return structures


def load_npy_structure(folders, cancel_event=None):
    if os.path.exists(os.path.join(folders, "type.raw")):
        return _load_npy_structure(folders, cancel_event=cancel_event)
    else:
        structures = []
        if os.path.isdir(folders):
            for folder in Path(folders).iterdir():
                if cancel_event is not None and getattr(cancel_event, "is_set", None) and cancel_event.is_set():
                    break
                structures.extend(load_npy_structure(folder, cancel_event=cancel_event))

        return structures

@utils.timeit
def save_npy_structure(folder:str, structures:list[Structure]):
    """
    保存结构信息到指定的文件夹，根据Config_type将数据组织
    :param folder: 保存的目标文件夹
    :param structures: 包含结构信息的Structure列表
    """


    # 确保文件夹存在，如果不存在则创建
    if not os.path.exists(folder):
        os.makedirs(folder)
    # 创建用于保存数据的字典

    dataset_dict = defaultdict(lambda: defaultdict(list))

    # 遍历所有结构并收集数据
    for structure in structures:
        # 从结构对象中提取数据
        config_type=structure.tag
        dataset_dict[config_type]["box"].append(structure.lattice.flatten())  # 确保box是3x3矩阵展平为1D数组
        dataset_dict[config_type]["coord"].append(structure.atomic_properties["pos"].flatten())  # 确保coords是1D数组
        dataset_dict[config_type]["species"].append(structure.atomic_properties["species"])

        # 保存每个额外字段（如果有）
        for prop_info  in structure.properties:
            name=prop_info["name"]
            if name not in [  "species", "pos"]:
                dataset_dict[config_type][name].append(structure.atomic_properties[name].flatten())
        if "virial" in structure.additional_fields:
            virial = structure.additional_fields["virial"]
            dataset_dict[config_type]["virial"].append(virial)
        if "energy" in structure.additional_fields:
            dataset_dict[config_type]["energy"].append(structure.energy)


    for config ,data in dataset_dict.items():
        save_path = os.path.join(folder, config,"set.000")
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        species=data["species"][0]
        unique_species = list(set([species   for species in species]))
        np.savetxt(os.path.join(folder,config,  "type_map.raw"), unique_species, fmt="%s")
        type_data = np.array([unique_species.index(species)   for species in species]).flatten()
        np.savetxt(os.path.join(folder, config, "type.raw"), type_data, fmt="%d")
        # 保存额外字段（如果有）
        for key, value in data.items():
            if key =="species":
                continue


            np.save(os.path.join(save_path, f"{key}.npy"), np.vstack(value ))





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2025/1/1 15:06
# @Author  : 兵
# @email    : 1747193328@qq.com
import os
os.environ["VISPY_IGNORE_OLD_VERSION"] = "true"

# os.environ["VISPY_PYQT5_SHARE_CONTEXT"] = "true"

import numpy as np

from PySide6.QtGui import QBrush, QColor, QPen
from vispy import scene

from vispy.color import ColorArray
from vispy.visuals.filters import MarkerPickingFilter
from NepTrainKit import utils
from NepTrainKit.config import Config
from NepTrainKit.core.canvas.base.canvas import VispyCanvasLayoutBase
from NepTrainKit.core.io import NepTrainResultData
from NepTrainKit.core.types import Brushes, Pens



class ViewBoxWidget(scene.Widget):

    def __init__(self, title, *args, **kwargs):
        super(ViewBoxWidget, self).__init__(*args, **kwargs)

        self.unfreeze()
        self.grid = self.add_grid(margin=0)

        self.grid.spacing = 0
        self.title_label = scene.Label(title, color='black',font_size=8)
        self.title_label.height_max = 30
        self.grid.add_widget(self.title_label, row=0, col=0, col_span=3)

        self.yaxis = scene.AxisWidget(orientation='left',
                                 axis_width=1,
                                 # axis_label='Y Axis',
                                 # axis_font_size=12,
                                 # axis_label_margin=10,
                                 tick_label_margin=5,
                                 axis_color="black",
                                 text_color="black"
                                 )
        self.yaxis.width_max = 50
        self.grid.add_widget(self.yaxis, row=1, col=0)

        self.xaxis = scene.AxisWidget(orientation='bottom',
                                 axis_width=1,
                                 tick_label_margin=10,
                                 axis_color="black",
                                 text_color="black"

                                 )

        self.xaxis.height_max = 30
        self.grid.add_widget(self.xaxis, row=2, col=1)

        right_padding = self.grid.add_widget(row=1, col=2, row_span=1)
        right_padding.width_max = 5
        self._view = self.grid.add_view(row=1, col=1,  )

        self._view.camera = scene.cameras.PanZoomCamera()
        self._view.camera.interactive = False

        self.xaxis.link_view(self._view)
        self.yaxis.link_view(self._view)

        self.text=  scene.Text('', parent=self._view.scene, color='red',anchor_x="left", anchor_y="top" )
        self.text.font_size = 8


        self.data=np.array([])

        # Configurable marker antialias and size
        try:
            self.marker_antialias = Config.getfloat("widget", "vispy_marker_antialias", 0.5)
        except Exception:
            self.marker_antialias = 0.5
        try:
            self.marker_size_default = Config.getint("widget", "vispy_marker_size", 6)
        except Exception:
            self.marker_size_default = 6

        self.picking_filter = MarkerPickingFilter()

        self._scatter=None
        # Overlay marker layers by name (e.g., 'selected', 'show')
        self._overlays = {}

        self._diagonal=None
        self.current_point=None
        self.freeze()



    def convert_color(self, obj):
        if isinstance(obj, (QPen, QBrush)):

            color = obj.color()
            edge_color = list(color.getRgbF())
        elif isinstance(obj, QColor):
            color = obj
            edge_color = list(color.getRgbF())

        else:
            edge_color = obj

        return edge_color

    def auto_range(self):
        if self._scatter is None:
            return

        pos = self._scatter._data["a_position"]
        if pos.size==0:
            return
        x_range = [10000, -10000]
        y_range = [10000, -10000]


        x = pos[:,0]
        y = pos[:,1]

        x = x[x > -10000]
        y = y[y > -10000]
        if x.size == 0:
            x_range =[0,1]
            y_range =[0,1]

        else:

            x_min = np.min(x)
            x_max = np.max(x)
            y_min = np.min(y)
            y_max = np.max(y)
            if x_min < x_range[0]:
                x_range[0] = x_min
            if x_max > x_range[1]:
                x_range[1] = x_max
            if y_min < y_range[0]:
                y_range[0] = y_min
            if y_max > y_range[1]:
                y_range[1] = y_max
        # self._view.camera.set_range( )
        #
        if "descriptor"!=self.title:
            real_range=(min(x_range[0],y_range[0]),max(x_range[1],y_range[1]))
            # Provide z-range to avoid VisPy querying scene bounds for z (empty visuals would error)
            self._view.camera.set_range(x=real_range, y=real_range, z=(0, 0))
        else:
            self._view.camera.set_range(x=x_range, y=y_range, z=(0, 0))

    def set_current_point(self, x,y):

        if np.array(x).size == 0:
            if self.current_point is not None:
                self.current_point.parent=None
                self.current_point=None

            return
        if self.current_point is None:
            # Create a top-most marker layer for current structure
            self.current_point = scene.visuals.Markers(antialias=1)
            # Ensure it renders above all other markers/overlays
            self.current_point.order = 100
            # Disable depth testing so nothing can occlude it
            self.current_point.update_gl_state(depth_test=False)
            self._view.add(self.current_point)

        z=np.full(x.shape,2)
        current_size = Config.getint("plot", "current_marker_size", 20) or 20
        self.current_point.set_data(
            np.vstack([x, y, z]).T,
            face_color=self.convert_color(Brushes.Current),
            edge_color=self.convert_color(Pens.Current),
            symbol='star',
            size=current_size,
        )

    def scatter(self,x,y,data,brush=None,pen=None ,**kwargs):
        if self._scatter is None:
            # Use configurable antialias level (default 0.5)
            self._scatter = scene.visuals.Markers(antialias=self.marker_antialias)
            self._scatter.order=1
            self._scatter.attach(self.picking_filter)

            self._view.add(self._scatter)

        self.data=data

        if brush is not None:

            kwargs["face_color"]=self.convert_color(brush)
        if pen is not None:

            kwargs["edge_color"]=self.convert_color(pen)
        if x.size != 0:

            pos = np.vstack([x, y], dtype=np.float32).T
            # Ensure a default size if caller didn't provide one
            if 'size' not in kwargs or kwargs.get('size') is None:
                kwargs['size'] = self.marker_size_default
            # self._scatter.update_gl_state(depth_test=False)
            self._scatter.set_data(pos, **kwargs)
            self.auto_range()
        else:
            self._scatter.set_data(np.empty((0, 3)), **kwargs)
            self.auto_range()
        self.update_diagonal()
        return self._scatter

    def line(self,x,y,**kwargs):
        xy=np.vstack([x,y]).T

        line=scene.Line(xy , **kwargs)
        self.view.add(line)
        return line
    def add_diagonal(self,**kwargs):


        x_domain = self.xaxis.axis.domain
        line_data = np.linspace(*x_domain,num=100)
        self._diagonal=self.line(line_data,line_data,**kwargs)

        self._diagonal.order=3
    def update_diagonal(self):
        if self._diagonal is None:
            return None

        x_domain = self.xaxis.axis.domain

        line_data = np.linspace(*x_domain,num=100)
        xy = np.vstack([line_data, line_data]).T
        self._diagonal.set_data(xy)

    @property
    def view(self):
        return self._view

    def _ensure_overlay(self, name:str, color, size:int=9, symbol:str='o'):
        if name in self._overlays and self._overlays[name] is not None:
            return self._overlays[name]
        ov = scene.visuals.Markers(antialias=1)
        ov.order = 4  # above base scatter and diagonal
        # keep same scene/camera
        self._view.add(ov)
        # overlays are 2D; no need for depth test
        ov.update_gl_state(depth_test=False)
        # initialize with empty data and hide from bounds
        ov.set_data(np.empty((0, 2), dtype=np.float32), face_color=self.convert_color(color), edge_width=0, symbol=symbol, size=size)
        ov.visible = False
        self._overlays[name] = ov
        return ov

    def set_overlay_positions(self, name:str, pos:np.ndarray, color=None, size:int=9, symbol:str='o'):
        """Replace an overlay layer's geometry in one call."""
        if pos is None:
            pos = np.empty((0, 2), dtype=np.float32)
        ov = self._ensure_overlay(name, color=color if color is not None else Brushes.Selected, size=size, symbol=symbol)
        kwargs = {}
        if color is not None:
            kwargs['face_color'] = self.convert_color(color)
        # Use face fill for highlight; no edge for lower cost
        pos = np.asarray(pos, dtype=np.float32)
        ov.set_data(pos=pos, edge_width=0, symbol=symbol, size=size, **kwargs)
        ov.visible = bool(pos.size)
        return ov

    def clear_overlays(self):
        """Hide and clear all overlay layers for this view."""
        empty = np.empty((0, 2), dtype=np.float32)
        for ov in self._overlays.values():
            ov.set_data(pos=empty, edge_width=0, symbol='o', size=9)
            ov.visible = False

    @property
    def title(self):
        return self.title_label._text_visual.text
    @property
    def rmse_size(self):
        return self.text.font_size
    @rmse_size.setter
    def rmse_size(self,size):

        self.text.font_size=size
    @title.setter
    def title(self, t):

        if t==self.title:
            return
        self.title_label._text_visual.text = t
        if t != "descriptor":
            self.add_diagonal(color="red", width=3, antialias=True, method='gl')
class CombinedMeta(type(VispyCanvasLayoutBase), type(scene.SceneCanvas) ):
    pass


class VispyCanvas(VispyCanvasLayoutBase, scene.SceneCanvas, metaclass=CombinedMeta):

    def __init__(self, *args, **kwargs):

        VispyCanvasLayoutBase.__init__(self)

        scene.SceneCanvas.__init__(self, *args,    **kwargs)

        self.unfreeze()
        self.nep_result_data = None

        # Per-axes overlay state: track indices to render in overlays without touching base VBO
        self._selected_by_plot = {}
        self._show_by_plot = {}


        self.grid = self.central_widget.add_grid(margin=0, spacing=0)
        self.grid.spacing = 0


        self.events.mouse_double_click.connect(self.switch_view_box)
        self.path_line = scene.visuals.Line(color='red', method='gl', antialias=False)
        self._path_update_step = 3
        # Use filters to affect the rendering of the mesh.

    def clear_axes(self):
        for widget in self.axes_list:
            widget._stretch = (None, None)
            widget.parent=None
            self.grid.remove_widget(widget)

        super().clear_axes()


    def set_nep_result_data(self,dataset):
        self.nep_result_data:NepTrainResultData=dataset


    def point_at(self,pos):
        if self.nep_result_data is None:
            return None
        current_axes=self._get_clicked_axes(pos)
        if current_axes is None:
            return None
        # adjust the event position for hidpi screens
        render_size = tuple(int(round(d * self.pixel_scale)) for d in self.size)
        x_pos = int(round(pos[0] * self.pixel_scale))
        y_pos = int(round(render_size[1] - (pos[1] * self.pixel_scale)))
        # print(canvas.pixel_scale)
        # render a small patch around the mouse cursor
        restore_state = not current_axes.picking_filter.enabled
        current_axes.picking_filter.enabled = True
        # Temporarily hide overlays and current-point so picking only sees base scatter
        hidden = []
        if hasattr(current_axes, '_overlays'):
            for ov in current_axes._overlays.values():
                if ov is not None and ov.visible:
                    hidden.append(ov)
                    ov.visible = False
        if getattr(current_axes, 'current_point', None) is not None and current_axes.current_point.visible:
            hidden.append(current_axes.current_point)
            current_axes.current_point.visible = False
        # Also hide drawing helpers to avoid contaminating the pick render
        path_was_visible = False
        if getattr(self, 'path_line', None) is not None and self.path_line.parent is not None:
            path_was_visible = getattr(self.path_line, 'visible', True)
            self.path_line.visible = False
        diag_was_visible = False
        if getattr(current_axes, '_diagonal', None) is not None:
            diag_was_visible = getattr(current_axes._diagonal, 'visible', False)
            current_axes._diagonal.visible = False
        current_axes._scatter.update_gl_state(blend=False)
        picking_render = self.render(
            crop=(x_pos - 3, y_pos - 3, 7, 7),
            bgcolor=(0, 0, 0, 0),
            alpha=True,
        )
        # Restore previously hidden visuals
        for v in hidden:
            v.visible = True
        if getattr(self, 'path_line', None) is not None and self.path_line.parent is not None:
            self.path_line.visible = path_was_visible
        if getattr(current_axes, '_diagonal', None) is not None:
            current_axes._diagonal.visible = diag_was_visible
        if restore_state:
            current_axes.picking_filter.enabled = False
        current_axes._scatter.update_gl_state(blend=not current_axes.picking_filter.enabled)

        # unpack indices in patch and pick the nearest valid marker index
        patch = (picking_render.view(np.uint32) - 1)[:, :, 0]
        if current_axes.data.size == 0:
            return None
        # valid indices are in [0, data_size)
        valid_mask = patch < int(current_axes.data.size)
        if not np.any(valid_mask):
            return None
        # compute distance to center for valid pixels and choose nearest
        h, w = patch.shape
        yy, xx = np.mgrid[0:h, 0:w]
        cy = (h - 1) / 2.0
        cx = (w - 1) / 2.0
        dist2 = (yy - cy) ** 2 + (xx - cx) ** 2
        # set invalid distances to large
        dist2 = np.where(valid_mask, dist2, 1e9)
        iy, ix = np.unravel_index(np.argmin(dist2), dist2.shape)
        marker_idx = int(patch[iy, ix])
        return marker_idx
    def on_mouse_press(self, event):
        """鼠标按下事件，开始记录轨迹"""

        if not self.draw_mode:

            index = self.point_at(event.pos)

            current_axes = self._get_clicked_axes(event.pos)

            if index is not None:
                structure_index=current_axes.data[index]


                self.structureIndexChanged.emit(structure_index)

            return False

        if event.button == 1 or event.button ==2:
            if self.draw_mode:

                tr = self.scene.node_transform(self.current_axes.view.scene)
                x, y, _, _ = tr.map(event.pos)
                self.mouse_path = [[x, y]]  # 记录初始点
                self.path_line.set_data(pos=np.array(self.mouse_path))  # 初始化轨迹线
                self.current_axes.view.add(self.path_line)


    def on_mouse_move(self, event):
        """鼠标移动事件，动态记录轨迹"""

        if not self.draw_mode:
            return
        if (event.button == 1 or event.button ==2) and len(self.mouse_path) > 0:
            # 记录当前鼠标位置
            tr = self.scene.node_transform(self.current_axes.view.scene)
            x, y, _, _ = tr.map(event.pos)

            self.mouse_path.append([x,y])

            # 更新轨迹线
            if (len(self.mouse_path) % getattr(self, '_path_update_step', 3)) == 0:
                self.path_line.set_data(pos=np.array(self.mouse_path))

    def on_mouse_release(self, event):
        """鼠标释放事件，结束记录"""
        if not self.draw_mode:
            return
        if event.button == 1 or event.button ==2:
            # 停止记录
            reverse=event.button == 2

            self.path_line.parent=None

            if len(self.mouse_path)>2:

                self.select_point_from_polygon(np.array(self.mouse_path),reverse)
            else:
                # 右键的话  选中单个点
                index = self.point_at(event.pos)
                if index is not None:
                    structure_index = self.current_axes.data[index]

                    self.select_index(structure_index,reverse)

            self.mouse_path = []

    def _get_clicked_axes(self,pos):
        view = self.visual_at(pos)

        if isinstance(view, scene.ViewBox)  :
            for axes in self.axes_list:
                if axes.view == view:
                    return axes

        return None
    def switch_view_box(self,event ):
        mouse_pos = event.pos
        axes=self._get_clicked_axes(mouse_pos)
        if axes is None:
            return
        if self.current_axes != axes:
            self.set_current_axes(axes)
            self.set_view_layout()

    def init_axes(self,axes_num   ):
        self.clear_axes()
        for r in range(axes_num):
            plot = ViewBoxWidget(title="")
            self.axes_list.append(plot)
        self.set_view_layout()
        self.update()

    def set_view_layout(self):
        if len(self.axes_list)==0:
            return
        if self.current_axes not in self.axes_list:
            self.set_current_axes(self.axes_list[0])
            return

        i = 0
        row_0_col_span=len(self.axes_list)-1
        for widget in self.axes_list:
            widget._stretch = (None, None)
            self.grid.remove_widget(widget)

            if widget == self.current_axes:
                widget.rmse_size=8
                self.grid.add_widget(widget, row=0, col=0, row_span=6, col_span=row_0_col_span)
            else:
                widget.rmse_size=4

                self.grid.add_widget(widget, row=6, col=i, row_span=2, col_span=1)

                i += 1

    def auto_range(self):
        self.current_axes.auto_range()


    def pan(self ,checked):
        self.current_axes.view.camera.interactive = checked



    def pen(self, checked):
        if self.current_axes is None:
            return False

        if checked:
            self.draw_mode = True
            # 初始化鼠标状态和轨迹数据

        else:
            self.draw_mode = False
            pass

    @utils.timeit

    def plot_nep_result(self):
        self.nep_result_data.select_index.clear()
        # Clear all overlays so deleted selections do not persist visually
        for plot in self.axes_list:
            if hasattr(plot, 'clear_overlays'):
                plot.clear_overlays()
            if plot in self._selected_by_plot:
                self._selected_by_plot[plot].clear()
            if plot in self._show_by_plot:
                self._show_by_plot[plot].clear()

        for index,_dataset in enumerate(self.nep_result_data.datasets):

            plot=self.axes_list[index]


            plot.title= _dataset.title

            marker_size = Config.getint("widget", "vispy_marker_size", 6) or 6
            plot.scatter(_dataset.x,
                         _dataset.y,
                         data=_dataset.structure_index,
                         brush=Brushes.get(_dataset.title.upper()) ,
                         pen=Pens.get(_dataset.title.upper()),
                         symbol='o',
                         size=marker_size,

                                      )

            # continue
            if _dataset.group_array.num !=0:
                #更新结构
                if self.structure_index not in _dataset.group_array.now_data:
                    self.structure_index=_dataset.group_array.now_data[0]
                    self.structureIndexChanged.emit(self.structure_index)

            else:
                plot.set_current_point([], [])

            if _dataset.title not in ["descriptor"]:
            #
                pos=self.convert_pos(plot,(0.1 ,0.8))
                text=f"rmse: {_dataset.get_formart_rmse()}"
                plot.text.text=text
                plot.text.pos=pos

    def convert_pos(self,plot,pos):
        x_range = plot.xaxis.axis.domain  # x轴范围 [xmin, xmax]
        y_range = plot.yaxis.axis.domain # y轴范围 [ymin, ymax]

        # 将百分比位置转换为坐标
        x_percent = pos[0] # 50% 对应 x 轴中间
        y_percent =  pos[1]  # 20% 对应 y 轴上的某个位置

        x_pos = x_range[0] + x_percent * (x_range[1] - x_range[0])  # 根据百分比计算实际位置
        y_pos = y_range[0] + y_percent * (y_range[1] - y_range[0])  # 根据百分比计算实际位置
        return x_pos,y_pos
    def plot_current_point(self,structure_index):
        self.structure_index=structure_index
        for plot in  self.axes_list :
            dataset=self.get_axes_dataset(plot)
            array_index=dataset.convert_index(structure_index)
            if dataset.is_visible(array_index) :

                data=dataset.all_data[array_index,: ]
                plot.set_current_point(data[:,dataset.x_cols].flatten(),
                                       data[:, dataset.y_cols].flatten(),
                                       )
            else:
                plot.set_current_point([], [])
    @utils.timeit
    def update_scatter_color(self,structure_index,color=Brushes.Selected):
        # Switch to overlay layers so we don't reupload the entire base VBO
        idx = np.atleast_1d(np.asarray(structure_index)).astype(np.int64)
        if idx.size == 0:
            return

        for plot in self.axes_list:
            if not plot._scatter:
                continue
            # init overlay sets for this plot
            if plot not in self._selected_by_plot:
                self._selected_by_plot[plot] = set()
            if plot not in self._show_by_plot:
                self._show_by_plot[plot] = set()

            if color is Brushes.Default:
                # remove from both overlays
                self._selected_by_plot[plot].difference_update(idx.tolist())
                self._show_by_plot[plot].difference_update(idx.tolist())
            elif color is Brushes.Selected:
                # add to selected, remove from show to avoid duplicates
                self._selected_by_plot[plot].update(idx.tolist())
                self._show_by_plot[plot].difference_update(idx.tolist())
            elif color is Brushes.Show:
                self._show_by_plot[plot].update(idx.tolist())
            else:
                # Fallback: treat as selected
                self._selected_by_plot[plot].update(idx.tolist())

            # Recompute overlay positions lazily from dataset (handles decimation of base)
            dataset = self.get_axes_dataset(plot)
            if dataset is None:
                continue

            def _indices_to_positions(indices:set[int]):
                if not indices:
                    return np.empty((0, 2), dtype=np.float32)
                indices_arr = np.fromiter(indices, dtype=np.int64)
                # Only consider currently visible (active) structures
                # The scatter uses flattened x/y; structure_index aligns with those
                try:
                    sidx = dataset.structure_index
                    mask = np.isin(sidx, indices_arr)
                    if not np.any(mask):
                        return np.empty((0, 2), dtype=np.float32)
                    x = dataset.x[mask]
                    y = dataset.y[mask]
                    return np.vstack([x, y], dtype=np.float32).T
                except Exception:
                    return np.empty((0, 2), dtype=np.float32)

            sel_pos = _indices_to_positions(self._selected_by_plot[plot])
            show_pos = _indices_to_positions(self._show_by_plot[plot])

            # Update overlays (filled squares, no edges for perf)
            overlay_size = Config.getint("widget", "vispy_marker_size", 6) or 6
            if sel_pos.size:
                plot.set_overlay_positions('selected', sel_pos, color=Brushes.Selected, size=overlay_size, symbol='o')
            else:
                plot.set_overlay_positions('selected', np.empty((0, 2), dtype=np.float32), color=Brushes.Selected, size=overlay_size, symbol='o')

            if show_pos.size:
                plot.set_overlay_positions('show', show_pos, color=Brushes.Show, size=overlay_size, symbol='o')
            else:
                plot.set_overlay_positions('show', np.empty((0, 2), dtype=np.float32), color=Brushes.Show, size=overlay_size, symbol='o')


    def select_point_from_polygon(self,polygon_xy,reverse ):
        index=self.is_point_in_polygon(np.column_stack([self.current_axes._scatter._data["a_position"][:,0],self.current_axes._scatter._data["a_position"][:,1]]),polygon_xy)
        index = np.where(index)[0]
        select_index=self.current_axes.data[index].tolist()
        self.select_index(select_index,reverse)



from typing import Union

from PySide6.QtCore import QRectF, Qt, QSize, Signal
from PySide6.QtGui import QPainter, QIcon, QPainterPath, QColor
from PySide6.QtWidgets import QApplication, QVBoxLayout, QWidget, QHBoxLayout, QSizePolicy
from qfluentwidgets import (PushButton, TransparentPushButton, FluentIconBase,
                            FlowLayout, TransparentToolButton, FluentIcon, TransparentTogglePushButton)

from qfluentwidgets.common.overload import singledispatchmethod



class CloseWidgetBase(QWidget):
    """ Split widget base class """

    closeClicked = Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.isHover=False
        self.isPressed=False
        self.closeButton = TransparentToolButton( self)
        self.closeButton.setIcon(FluentIcon.CLOSE)
        self.hBoxLayout = QHBoxLayout(self)
        self.hBoxLayout.setSpacing(0)
        self.hBoxLayout.setContentsMargins(0, 0, 0, 0)
        self.hBoxLayout.addWidget(self.closeButton)

        self.closeButton.clicked.connect(self.closeClicked)
        self.borderRadius=3
        self.backgroundColor="#FFFFFF"
        self.setBackgroundColor(self.backgroundColor)
        self.setAttribute(Qt.WA_TranslucentBackground)
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)


    def setWidget(self, widget: QWidget):
        """ set the widget on left side """
        self.hBoxLayout.insertWidget(0, widget, 1, Qt.AlignLeft)

    def setDropButton(self, button):
        """ set drop dow button """
        self.hBoxLayout.removeWidget(self.closeButton)
        self.closeButton.deleteLater()

        self.closeButton = button
        self.closeButton.clicked.connect(self.closeClicked)

        self.hBoxLayout.addWidget(button)
    def setBackgroundColor(self,color):
        self.backgroundColor= color
        self.update()
    def setWidget(self, widget: QWidget):
        """ set the widget on left side """
        self.hBoxLayout.insertWidget(0, widget, 1, Qt.AlignLeft)
 
    def paintEvent(self, e):
        painter = QPainter(self)
        painter.setRenderHints(QPainter.Antialiasing)

        w, h = self.width(), self.height()
        r = self.borderRadius
        d = 2 * r

        isDark = False

        # draw top border
        path = QPainterPath()
        # path.moveTo(1, h - r)
        path.arcMoveTo(1, h - d - 1, d, d, 240)
        path.arcTo(1, h - d - 1, d, d, 225, -60)
        path.lineTo(1, r)
        path.arcTo(1, 1, d, d, -180, -90)
        path.lineTo(w - r, 1)
        path.arcTo(w - d - 1, 1, d, d, 90, -90)
        path.lineTo(w - 1, h - r)
        path.arcTo(w - d - 1, h - d - 1, d, d, 0, -60)

        topBorderColor = QColor(0, 0, 0, 20)
        if isDark:
            if self.isPressed:
                topBorderColor = QColor(255, 255, 255, 18)
            elif self.isHover:
                topBorderColor = QColor(255, 255, 255, 13)
        else:
            topBorderColor = QColor(0, 0, 0, 15)

        painter.strokePath(path, topBorderColor)

        # draw bottom border
        path = QPainterPath()
        path.arcMoveTo(1, h - d - 1, d, d, 240)
        path.arcTo(1, h - d - 1, d, d, 240, 30)
        path.lineTo(w - r - 1, h - 1)
        path.arcTo(w - d - 1, h - d - 1, d, d, 270, 30)

        bottomBorderColor = topBorderColor
        if not isDark and self.isHover and not self.isPressed:
            bottomBorderColor = QColor(0, 0, 0, 27)

        painter.strokePath(path, bottomBorderColor)

        # draw background
        painter.setPen(Qt.NoPen)
        rect = self.rect().adjusted(1, 1, -1, -1)
        painter.setBrush(self.backgroundColor)
        painter.drawRoundedRect(rect, r, r)
 

class TagPushButton(CloseWidgetBase):
    checkedSignal = Signal(str)
    @singledispatchmethod
    def __init__(self, parent: QWidget = None,checkable: bool = False):

        super(TagPushButton, self).__init__(parent)
        if checkable:

            self.button = TransparentTogglePushButton(FluentIcon.TAG,"", self)
            self.button.toggled.connect(lambda :self.checkedSignal.emit(self.button.text()))
        else:
            self.button = TransparentPushButton(FluentIcon.TAG, "", self)
        self.button.setObjectName('PushButton')

        self.setWidget(self.button)


    @__init__.register
    def _(self, text: str, parent: QWidget = None, icon: Union[QIcon, str, FluentIconBase] = None, checkable: bool = False):
        self.__init__(parent,checkable)
        self.setText(text)

        self.setIcon(icon)

    @__init__.register
    def _(self, icon: QIcon, text: str, checkable: bool = False, parent: QWidget = None):
        self.__init__(text, parent, icon,checkable)

    @__init__.register
    def _(self, icon: FluentIconBase, text: str, checkable: bool = False, parent: QWidget = None):
        self.__init__(text, parent, icon,checkable)
    def text(self):
        return self.button.text()

    def setText(self, text: str):
        self.button.setText(text)
        self.adjustSize()

    def icon(self):
        return self.button.icon()

    def setIcon(self, icon: Union[QIcon, FluentIconBase, str]):
        self.button.setIcon(icon)

    def setIconSize(self, size: QSize):
        self.button.setIconSize(size)

class TagGroup(QWidget):
    tagRemovedSignal = Signal(str)
    tagCheckedSignal = Signal(str)
    def __init__(self,tags=None, parent: QWidget = None):
        super(TagGroup, self).__init__(parent)
        self._layout = FlowLayout(self, needAni=True)
        self.tags={}

        if tags is not None:
            for tag in tags:
                self.add_tag(tag)

    def has_tag(self, tag):
        return tag in self.tags

    def add_tag(self, tag,color=None,icon=FluentIcon.TAG,checkable=False) ->TagPushButton:


        button = TagPushButton(icon,tag,checkable, self)
        button.checkedSignal.connect(self.tagCheckedSignal)
        button.closeClicked.connect(lambda _tag=tag:self.del_tag(_tag))
        if color is not None:
            button.setBackgroundColor(color)
        self._layout.addWidget(button)
        self.tags[tag]=button
        return button

    def del_tag(self, tag):
        button = self.tags[tag]
        self._layout.removeWidget(button)
        button.deleteLater()
        self.tagRemovedSignal.emit(tag)
        del self.tags[tag]
        self._layout.update()
if __name__ == '__main__':
    import sys
    app = QApplication(sys.argv)
    from PySide6.QtWidgets import QFrame

    frame = QWidget()
    frame.resize(800, 600)
    layout = QVBoxLayout()
    frame.setLayout(layout)
    pushButton = TagPushButton(FluentIcon.PROJECTOR,"dwasd",frame )
    layout.addWidget(pushButton)

    frame.show()
    app.exec_()


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/20 22:22
# @Author  : 兵
# @email    : 1747193328@qq.com
import time
import json

from NepTrainKit.core.calculator import NEPProcess, NepCalculator

start = time.time()
import numpy as np
from PySide6.QtWidgets import QHBoxLayout, QWidget, QProgressDialog


from NepTrainKit import utils
from NepTrainKit.core import MessageManager
from NepTrainKit.config import Config

from NepTrainKit.custom_widget import (
    GetIntMessageBox,
    SparseMessageBox,
    IndexSelectMessageBox,
    RangeSelectMessageBox,
    EditInfoMessageBox,
    ShiftEnergyMessageBox,
    DFTD3MessageBox,
)
from NepTrainKit.core.types import NepBackend, SearchType, CanvasMode
from NepTrainKit.core.io.select import farthest_point_sampling
from NepTrainKit.views.toolbar import NepDisplayGraphicsToolBar
from NepTrainKit.core.energy_shift import shift_dataset_energy, suggest_group_patterns


class NepResultPlotWidget(QWidget):
    def __init__(self,parent=None):
        super().__init__(parent)
        self._parent=parent
        self.tool_bar: NepDisplayGraphicsToolBar
        self.draw_mode=False
        # self.setRenderHint(QPainter.Antialiasing, False)
        self._layout = QHBoxLayout(self)
        self.setLayout(self._layout)
        canvas_type = Config.get("widget","canvas_type",CanvasMode.PYQTGRAPH)

        self.last_figure_num=None
        self.swith_canvas(canvas_type)

    def swith_canvas(self,canvas_type:CanvasMode="pyqtgraph"):

        if canvas_type == CanvasMode.PYQTGRAPH:
            from NepTrainKit.core.canvas.pyqtgraph import PyqtgraphCanvas
            self.canvas = PyqtgraphCanvas(self)
            self._layout.addWidget(self.canvas)

        elif canvas_type == CanvasMode.VISPY:


            from NepTrainKit.core.canvas.vispy import VispyCanvas


            self.canvas = VispyCanvas(parent=self, bgcolor='white')
            self._layout.addWidget(self.canvas.native)
            # self.window().windowHandle().screenChanged.connect(self.canvas.native.screen_changed)



    # def clear(self):
    #     self.canvas.clear_axes()
        # self.last_figure_num=None

    def set_tool_bar(self, tool):
        self.tool_bar: NepDisplayGraphicsToolBar = tool
        self.tool_bar.panSignal.connect(self.canvas.pan)
        self.tool_bar.resetSignal.connect(self.canvas.auto_range)
        self.tool_bar.deleteSignal.connect(self.canvas.delete)
        self.tool_bar.revokeSignal.connect(self.canvas.revoke)
        self.tool_bar.penSignal.connect(self.canvas.pen)
        self.tool_bar.exportSignal.connect(self.export_descriptor_data)
        self.tool_bar.findMaxSignal.connect(self.find_max_error_point)
        self.tool_bar.discoverySignal.connect(self.find_non_physical_structures)
        self.tool_bar.sparseSignal.connect(self.sparse_point)
        self.tool_bar.shiftEnergySignal.connect(self.shift_energy_baseline)
        self.tool_bar.inverseSignal.connect(self.inverse_select)
        self.tool_bar.selectIndexSignal.connect(self.select_by_index)
        self.tool_bar.rangeSignal.connect(self.select_by_range)
        self.tool_bar.dftd3Signal.connect(self.calc_dft_d3)
        self.tool_bar.editInfoSignal.connect(self.edit_structure_info)
        self.canvas.tool_bar = self.tool_bar


    def __find_non_physical_structures(self):
        """
        对每个结构进行非物理距离判断
        """
        structure_list = self.canvas.nep_result_data.structure.now_data
        group_array = self.canvas.nep_result_data.structure.group_array.now_data
        radius_coefficient_config = Config.getfloat("widget","radius_coefficient",0.7)
        unreasonable_index=[]
        for structure,index in zip(structure_list,group_array):

            if not structure.adjust_reasonable(radius_coefficient_config):

                unreasonable_index.append(index)
            yield 1
        self.canvas.select_index(unreasonable_index,False)


    def find_non_physical_structures(self):
        """
        非物理结构函数入口
        """
        if self.canvas.nep_result_data is None:
            return
        progress_diag = QProgressDialog(f"" ,"Cancel",0,self.canvas.nep_result_data.structure.num,self._parent)
        thread=utils.LoadingThread(self._parent,show_tip=False )
        progress_diag.setFixedSize(300, 100)
        progress_diag.setWindowTitle("Finding non-physical structures")
        thread.progressSignal.connect(progress_diag.setValue)
        thread.finished.connect(progress_diag.accept)
        progress_diag.canceled.connect(thread.stop_work)  # 用户取消时终止线程
        thread.start_work(self.__find_non_physical_structures)
        progress_diag.exec()

    def find_max_error_point(self):
        dataset = self.canvas.get_axes_dataset(self.canvas.current_axes)

        if dataset is None:
            return

        box= GetIntMessageBox(self._parent,"Please enter an integer N, it will find the top N structures with the largest errors")
        n = Config.getint("widget","max_error_value",10)
        box.intSpinBox.setValue(n)

        if not box.exec():
            return
        nmax= box.intSpinBox.value()
        Config.set("widget","max_error_value",nmax)
        index= (dataset.get_max_error_index(nmax))

        self.canvas.select_index(index,False)

    def sparse_point(self):
        if  self.canvas.nep_result_data is None:
            return
        box= SparseMessageBox(self._parent,"Please specify the maximum number of structures and minimum distance")
        n_samples = Config.getint("widget","sparse_num_value",10)
        distance = Config.getfloat("widget","sparse_distance_value",0.01)

        box.intSpinBox.setValue(n_samples)
        box.doubleSpinBox.setValue(distance)

        if not box.exec():
            return
        n_samples= box.intSpinBox.value()
        distance= box.doubleSpinBox.value()
        use_selection_region = bool(getattr(box, 'regionCheck', None) and box.regionCheck.isChecked())

        Config.set("widget","sparse_num_value",n_samples)
        Config.set("widget","sparse_distance_value",distance)

        dataset = self.canvas.nep_result_data.descriptor
        if dataset.now_data.size ==0:
            MessageManager.send_message_box("No descriptor data available","Error")
            return
        # Build region mask
        reverse=False

        points = dataset.now_data
        mask = np.ones(points.shape[0], dtype=bool)
        if use_selection_region:
            sel = np.asarray(list(self.canvas.nep_result_data.select_index), dtype=np.int64)
            if sel.size == 0:
                MessageManager.send_info_message("No selection found; FPS will run on full data.")
            else:
                # Map structure selection to descriptor rows via group_array
                struct_ids = dataset.group_array.now_data
                mask = np.isin(struct_ids, sel)
                if not np.any(mask):
                    MessageManager.send_info_message("Current selection has no points on this plot; FPS will run on full data.")
                    mask = np.ones(points.shape[0], dtype=bool)
                else:
                    reverse = True
                    MessageManager.send_info_message("When FPS sampling is performed in the designated area, the program will automatically deselect it, just click to delete!")

        if np.any(mask):
            subset = points[mask]
            idx_local = farthest_point_sampling(subset, n_samples=n_samples, min_dist=distance)
            # Map back to global row indices
            global_rows = np.where(mask)[0][np.asarray(idx_local, dtype=np.int64)]
        else:
            idx_local = []
            global_rows = np.array([], dtype=np.int64)
        # 获取所有索引（从 0 到 len(arr)-1）
        # all_indices = np.arange(dataset.now_data.shape[0])

        # 使用 setdiff1d 获取不在 indices_to_remove 中的索引
        # remove_indices = np.setdiff1d(all_indices, remaining_indices)
        structures = dataset.group_array[global_rows]
        self.canvas.select_index(structures.tolist(),reverse)

    def edit_structure_info(self):
        data = self.canvas.nep_result_data
        if data is None or len(data.select_index) == 0:
            MessageManager.send_info_message("No data selected!")
            return
        selected_structures = data.get_selected_structures()
        tags= {item for structure in selected_structures for item in structure.get_prop_key(True, True)}
        #这两个不允许删除
        tags.remove("species")
        tags.remove("pos")
        box = EditInfoMessageBox(self._parent)

        box.init_tags(list(tags))
        if not box.exec():
            return


        for structure in selected_structures:

            for remove_tag in box.remove_tag:
                if remove_tag in structure.additional_fields:
                    structure.additional_fields.pop(remove_tag)
                #这里是为了避免info和array有重复名字的 所以如果info有，就只删除info
                #如果在想删除array的 就要再操作一次即可

                elif remove_tag in structure.atomic_properties:
                    structure.remove_atomic_properties(remove_tag)

            for new_tag,value_text in box.new_tag_info.items():


                try:
                    value = json.loads(value_text)
                    if isinstance(value, list):
                        value = np.array(value)
                except Exception:
                    try:
                        value = float(value_text)
                    except Exception:
                        value = value_text
                structure.additional_fields[new_tag] = value
        MessageManager.send_info_message("Edit completed")

    def export_descriptor_data(self):
        if self.canvas.nep_result_data is None:
            MessageManager.send_info_message("NEP data has not been loaded yet!")
            return
        path = utils.call_path_dialog(self, "Choose a file save ", "file",default_filename="export_descriptor_data.out")
        if path:
            thread = utils.LoadingThread(self, show_tip=True, title="Exporting descriptor data")
            thread.start_work(self._export_descriptor_data, path)

    def _export_descriptor_data(self,path):

        if len(self.canvas.nep_result_data.select_index) == 0:
            MessageManager.send_info_message("No data selected!")
            return
        select_index=self.canvas.nep_result_data.descriptor.convert_index(list(self.canvas.nep_result_data.select_index))
        descriptor_data = self.canvas.nep_result_data.descriptor.now_data[select_index,:]
        if hasattr(self.canvas.nep_result_data,"energy") and self.canvas.nep_result_data.energy.num !=0:
            select_index = self.canvas.nep_result_data.energy.convert_index(
                list(self.canvas.nep_result_data.select_index))

            energy_data = self.canvas.nep_result_data.energy.now_data[select_index,1]
            descriptor_data = np.column_stack((descriptor_data,energy_data))

        with open(path, "w",encoding="utf8") as f:
            np.savetxt(f,descriptor_data,fmt='%.6g',delimiter='\t')


    def shift_energy_baseline(self):
        data = self.canvas.nep_result_data
        if data is None:
            return
        ref_index = list(data.select_index)
        # if len(ref_index) == 0:
        #     MessageManager.send_info_message("No data selected!")
        #     return

        max_generations = Config.getint("widget","max_generation_value",100000)
        population_size =  Config.getint("widget","population_size",40)
        convergence_tol = Config.getfloat("widget","convergence_tol", 1e-8)
        config_set = set(data.structure.get_all_config(SearchType.TAG))
        suggested = suggest_group_patterns(list(config_set))
        box = ShiftEnergyMessageBox(
            self._parent,
            "Specify regex groups for Config_type (comma separated)"
        )
        box.groupEdit.setText(";".join(suggested))
        box.genSpinBox.setValue(max_generations)

        box.sizeSpinBox.setValue(population_size)
        box.tolSpinBox.setValue(convergence_tol)


        if not box.exec():
            return

        pattern_text = box.groupEdit.text().strip()
        group_patterns = [p.strip() for p in pattern_text.split(';') if p.strip()]

        alignment_mode = box.modeCombo.currentText()


        max_generations = box.genSpinBox.value()
        population_size = box.sizeSpinBox.value()
        convergence_tol = box.tolSpinBox.value()
        Config.set("widget","max_generation_value",max_generations)
        Config.set("widget","population_size",population_size)
        Config.set("widget","convergence_tol",convergence_tol)
        config_set = set(data.structure.get_all_config(SearchType.TAG))
        progress_diag = QProgressDialog(f"", "Cancel", 0, len(config_set), self._parent)
        thread = utils.LoadingThread(self._parent, show_tip=False)
        progress_diag.setFixedSize(300, 100)
        progress_diag.setWindowTitle("Shift energies")
        thread.progressSignal.connect(progress_diag.setValue)
        thread.finished.connect(progress_diag.accept)
        progress_diag.canceled.connect(thread.stop_work)  # 用户取消时终止线程
        thread.start_work(
            shift_dataset_energy,
            structures=data.structure.now_data,
            reference_structures=data.structure.all_data[ref_index],
            max_generations=max_generations,
            population_size=population_size,
            convergence_tol=convergence_tol,
            group_patterns=group_patterns,
            alignment_mode=alignment_mode,
            nep_energy_array=data.energy.y,
        )
        progress_diag.exec()
        if hasattr(data, "energy") and data.energy.num != 0:
            for i, s in enumerate(data.structure.all_data):
                # print(s.per_atom_energy)
                data.energy.data._data[i, data.energy.x_cols] = s.per_atom_energy
        self.canvas.plot_nep_result()
    def _calc_dft_d3(self,mode,functional,cutoff,cutoff_cn):
        nep_result_data = self.canvas.nep_result_data
        nep_txt_path = nep_result_data.nep_txt_path



        #
        # if mode == 0:
        #     nep_calc = NepCalculator(
        #         model_file=nep_txt_path.as_posix(),
        #         backend=NepBackend(Config.get("nep", "backend", "auto")),
        #         batch_size=Config.getint("nep", "gpu_batch_size", 1000)
        #     )
        #     nep_potentials_array, nep_forces_array, nep_virials_array = nep_calc.calculate(nep_result_data.structure.now_data.tolist())
        #
        # elif mode == 2:
        #
        #     nep_calc = NepCalculator(
        #         model_file=nep_txt_path.as_posix(),
        #         backend=NepBackend.CPU,
        #         batch_size=Config.getint("nep", "gpu_batch_size", 1000)
        #     )
        #     nep_potentials_array, nep_forces_array, nep_virials_array = nep_calc.calculate_with_dftd3(
        #         nep_result_data.structure.now_data.tolist(),
        #         functional=functional,
        #         cutoff= cutoff,
        #         cutoff_cn= cutoff_cn
        #
        #     )
        #
        # else:
        nep_calc = NepCalculator(
            model_file=nep_txt_path.as_posix(),
            backend=NepBackend.CPU,
            batch_size=Config.getint("nep", "gpu_batch_size", 1000)
        )
        nep_potentials_array, nep_forces_array, nep_virials_array = nep_calc.calculate_dftd3(
            nep_result_data.structure.now_data.tolist(),
            functional=functional,
            cutoff=cutoff,
            cutoff_cn=cutoff_cn

        )
        now_atoms_num_list=nep_result_data.atoms_num_list[nep_result_data.structure.now_indices]
        split_indices = np.cumsum(now_atoms_num_list)[:-1]
        nep_forces_array = np.split(nep_forces_array, split_indices)
        nep_virials_array=nep_virials_array*now_atoms_num_list[:, np.newaxis]
        #
        # if mode < 3:
        #     for index, structure in enumerate(nep_result_data.structure.now_data):
        #         structure.energy = nep_potentials_array[index]
        #         structure.forces = nep_forces_array[index]
        #         structure.virial = nep_virials_array[index]
        # else:
        factor = 1 if mode == 0 else -1
        for index, structure in enumerate(nep_result_data.structure.now_data):
            structure.energy += nep_potentials_array[index] * factor
            structure.forces += nep_forces_array[index] * factor
            structure.virial += nep_virials_array[index] * factor

        now_indices = nep_result_data.structure.now_indices


        if hasattr(nep_result_data, "energy") and  nep_result_data.energy.num != 0:
            # print(s.per_atom_energy)
            ref_energies = np.array([s.per_atom_energy for s in nep_result_data.structure.now_data], dtype=np.float32).reshape(-1, 1)

            nep_result_data.energy.data._data[now_indices,  nep_result_data.energy.x_cols] = ref_energies
        if hasattr(nep_result_data, "force") and nep_result_data.force.num != 0:
            force_index=nep_result_data.force.convert_index(now_indices)
            ref_forces = np.vstack([s.forces for s in nep_result_data.structure.now_data], dtype=np.float32)

            nep_result_data.force.data._data[force_index,  nep_result_data.force.x_cols] = ref_forces
        if hasattr(nep_result_data, "virial") and nep_result_data.virial.num != 0:
            ref_virials = np.vstack([s.nep_virial for s in nep_result_data.structure.now_data], dtype=np.float32)
            # print(nep_result_data.structure.now_data[0].virial)
            # print(nep_result_data.structure.now_data[0].nep_virial)
            #
            # print(ref_virials[0])
            nep_result_data.virial.data._data[now_indices, nep_result_data.virial.x_cols] = ref_virials

            # print(nep_result_data.virial.data._data.tolist())
            if hasattr(nep_result_data, "stress") and nep_result_data.stress.num != 0:
                coefficient = (now_atoms_num_list / np.array(
                    [s.volume for s in nep_result_data.structure.now_data]))[:, np.newaxis]
                stress_array = ref_virials * coefficient * 160.21766208  # 单位转换\
                stress_array = stress_array.astype(np.float32)

                nep_result_data.stress.data._data[now_indices, nep_result_data.stress.x_cols] = stress_array



    def calc_dft_d3(self):


        if  self.canvas.nep_result_data is None:
            return

        function = Config.get("widget","functional","scan")
        cutoff = Config.getfloat("widget","cutoff",12)
        cutoff_cn = Config.getfloat("widget","cutoff_cn",6)
        mode = Config.getint("widget","d3_mode",0)

        box = DFTD3MessageBox(
            self._parent,
            "DFT D3"
        )
        box.functionEdit.setText(function)
        box.d1SpinBox.setValue(cutoff)
        box.d1cnSpinBox.setValue(cutoff_cn)
        box.modeCombo.setCurrentIndex(mode)
        if not box.exec():
            return

        mode = box.modeCombo.currentIndex()
        D3_cutoff = box.d1SpinBox.value()
        D3_cutoff_cn = box.d1cnSpinBox.value()
        functional=box.functionEdit.text().strip()
        Config.set("widget","cutoff",D3_cutoff)
        Config.set("widget","cutoff_cn",D3_cutoff_cn)
        Config.set("widget","functional",functional)
        Config.set("widget","d3_mode",mode)

        thread = utils.LoadingThread(self._parent, show_tip=True,title="calculating dftd3")

        thread.start_work(
            self._calc_dft_d3,
            mode,functional,cutoff,cutoff_cn
        )
        thread.finished.connect(self.canvas.plot_nep_result)

        # self.canvas.plot_nep_result()






    def inverse_select(self):
        self.canvas.inverse_select()

    def select_by_index(self):
        if self.canvas.nep_result_data is None:
            return
        box = IndexSelectMessageBox(self._parent, "Select structures by index")
        if not box.exec():
            return
        text = box.indexEdit.text().strip()
        use_origin = box.checkBox.isChecked()
        data = self.canvas.nep_result_data.structure
        total = data.all_data.shape[0] if use_origin else data.now_data.shape[0]
        indices = utils.parse_index_string(text, total)
        if not indices:
            return
        if not use_origin:
            indices = data.group_array.now_data[indices].tolist()
        self.canvas.select_index(indices, False)

    def select_by_range(self):
        if self.canvas.nep_result_data is None:
            return
        dataset = self.canvas.get_axes_dataset(self.canvas.current_axes)
        if dataset is None or dataset.now_data.size == 0:
            return
        box = RangeSelectMessageBox(self._parent, "Select structures by range")
        box.xMinSpin.setValue(float(np.min(dataset.x)))
        box.xMaxSpin.setValue(float(np.max(dataset.x)))
        box.yMinSpin.setValue(float(np.min(dataset.y)))
        box.yMaxSpin.setValue(float(np.max(dataset.y)))
        if not box.exec():
            return
        x_min, x_max = sorted([box.xMinSpin.value(), box.xMaxSpin.value()])
        y_min, y_max = sorted([box.yMinSpin.value(), box.yMaxSpin.value()])
        mask_x = (dataset.x >= x_min) & (dataset.x <= x_max)
        mask_y = (dataset.y >= y_min) & (dataset.y <= y_max)
        mask = mask_x & mask_y if box.logicCombo.currentText() == "AND" else mask_x | mask_y
        indices = np.unique(dataset.structure_index[mask]).tolist()
        self.canvas.select_index(indices, False)

    def set_dataset(self,dataset):

        if self.last_figure_num !=len(dataset.datasets):

            self.canvas.init_axes(len(dataset.datasets))
            self.last_figure_num = len(dataset.datasets)

        self.canvas.set_nep_result_data(dataset)
        self.canvas.plot_nep_result()


















#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/17 13:38
# @Author  : 兵
# @email    : 1747193328@qq.com

from PySide6.QtCore import QObject, Signal, Qt
from qfluentwidgets import   InfoBar, InfoBarIcon, InfoBarPosition,MessageBox

class MessageManager(QObject):
    """
    全局的消息弹窗 单例模式 直接导入调用即可
    比如
    from core import MessageManager
    MessageManager.send_info_message("这是一条info消息")
    MessageManager.send_info_message( "这是一条info消息","标题")

    """
    _instance = None
    showMessageSignal = Signal(InfoBarIcon, str, str)
    showBoxSignal= Signal(str, str)

    def __init__(self,parent=None):
        super().__init__()
        self._parent = parent
        self._instance:MessageManager
        self.showMessageSignal.connect(self._show_message)
        self.showBoxSignal.connect(self._show_box)

    @classmethod
    def _createInstance(cls,parent=None):
        # 创建实例
        if not cls._instance:
            cls._instance = MessageManager(parent)

    @classmethod
    def get_instance(cls):
        cls._createInstance()
        return cls._instance

    @classmethod
    def send_info_message(cls,message,title="Tip"):
        cls._createInstance()
        cls._instance.showMessageSignal.emit(InfoBarIcon.INFORMATION, message, title)

    @classmethod
    def send_success_message(cls,message,title="Success"):
        cls._createInstance()
        cls._instance.showMessageSignal.emit(InfoBarIcon.SUCCESS, message, title)

    @classmethod
    def send_warning_message(cls,message,title="Warning"):
        cls._createInstance()
        cls._instance.showMessageSignal.emit(InfoBarIcon.WARNING, message, title)

    @classmethod
    def send_error_message(cls, message,title="Error"):
        cls._createInstance()
        cls._instance.showMessageSignal.emit(InfoBarIcon.ERROR, message, title)

    @classmethod
    def send_message_box(cls,message,title="Tip"):
        cls._createInstance()
        cls._instance.showBoxSignal.emit(message, title)

    def _show_box(self,message,title ):
        w = MessageBox(title, message, self._parent)
        w.cancelButton.hide()
        w.exec_()

    def _show_message(self,msg_type,msg,title):

        if msg_type==InfoBarIcon.ERROR:
            duration=10000
        elif msg_type==InfoBarIcon.WARNING:
            duration=8000
        else:
            duration=3000
        InfoBar.new(msg_type,
            title=title,
            content=msg,
            orient=Qt.Orientation.Vertical,  # vertical layout
            isClosable=True,
            position=InfoBarPosition.TOP_RIGHT,
            duration=duration,
            parent=self._parent
        )






#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/18 13:26
# @Author  : 兵
# @email    : 1747193328@qq.com

import os
import traceback
from pathlib import Path
import numpy.typing as npt
import numpy as np
from loguru import logger
from NepTrainKit import module_path,utils
from NepTrainKit.core import MessageManager, Structure
from NepTrainKit.config import Config

from NepTrainKit.core.calculator import NEPProcess, run_nep_calculator, NepCalculator

from NepTrainKit.core.io.base import NepPlotData, StructureData, ResultData

from NepTrainKit.core.io.utils import read_nep_out_file, check_fullbatch, read_nep_in, parse_array_by_atomnum
from NepTrainKit.core.types import ForcesMode, NepBackend


class NepTrainResultData(ResultData):
    _energy_dataset: NepPlotData
    _force_dataset: NepPlotData
    _stress_dataset: NepPlotData
    _virial_dataset: NepPlotData
    def __init__(self,
                 nep_txt_path: Path|str,
                 data_xyz_path: Path|str,
                 energy_out_path: Path|str,
                 force_out_path: Path|str,
                 stress_out_path: Path|str,
                 virial_out_path: Path|str,
                 descriptor_path: Path|str

                 ):
        super().__init__(nep_txt_path,data_xyz_path,descriptor_path)
        self.energy_out_path = Path(energy_out_path)
        self.force_out_path = Path(force_out_path)
        self.stress_out_path = Path(stress_out_path)
        self.virial_out_path = Path(virial_out_path)

    @property
    def datasets(self):
        # return [self.energy, self.stress,self.virial, self.descriptor]
        return [self.energy,self.force,self.stress,self.virial, self.descriptor]

    @property
    def energy(self):
        return self._energy_dataset

    @property
    def force(self):
        return self._force_dataset

    @property
    def stress(self):
        return self._stress_dataset

    @property
    def virial(self):
        return self._virial_dataset

    @classmethod
    def from_path(cls, path ,model_type=0, *, structures: list[Structure] | None = None):
        dataset_path = Path(path)

        file_name=dataset_path.stem

        nep_txt_path = dataset_path.with_name(f"nep.txt")
        if not nep_txt_path.exists()  :
            nep89_path = os.path.join(module_path, "Config/nep89.txt")
            nep_txt_path=Path(nep89_path)
            MessageManager.send_warning_message(f"no find nep.txt; the program will use nep89 instead.")

        elif model_type>2:
            nep89_path = os.path.join(module_path, "Config/nep89.txt")
            nep_txt_path=Path(nep89_path)
            MessageManager.send_warning_message(f"NEPKit currently does not support model_type={model_type}; the program will use nep89 instead.")
        energy_out_path = dataset_path.with_name(f"energy_{file_name}.out")
        force_out_path = dataset_path.with_name(f"force_{file_name}.out")
        stress_out_path = dataset_path.with_name(f"stress_{file_name}.out")
        virial_out_path = dataset_path.with_name(f"virial_{file_name}.out")
        if file_name=="train":

            descriptor_path = dataset_path.with_name(f"descriptor.out")
        else:
            descriptor_path = dataset_path.with_name(f"descriptor_{file_name}.out")


        inst = cls(nep_txt_path,dataset_path,energy_out_path,force_out_path,stress_out_path,virial_out_path,descriptor_path)
        if structures is not None:
            try:
                inst.set_structures(structures)
            except Exception:
                pass
        return inst

    def _load_dataset(self) -> None:
        """加载或计算 NEP 数据集，并更新内部数据集属性。"""
        nep_in = read_nep_in(self.data_xyz_path.with_name("nep.in"))
        if self._should_recalculate(nep_in):
            energy_array, force_array, virial_array, stress_array = self._recalculate_and_save( )
        else:
            energy_array = read_nep_out_file(self.energy_out_path, dtype=np.float32,ndmin=2)
            force_array = read_nep_out_file(self.force_out_path, dtype=np.float32,ndmin=2)
            virial_array = read_nep_out_file(self.virial_out_path, dtype=np.float32,ndmin=2)
            stress_array = read_nep_out_file(self.stress_out_path, dtype=np.float32,ndmin=2)

            if energy_array.shape[0]!=self.atoms_num_list.shape[0]:
                self.energy_out_path.unlink(True)
                self.force_out_path.unlink(True)
                self.virial_out_path.unlink(True)
                self.stress_out_path.unlink(True)

                return self._load_dataset()


        self._energy_dataset = NepPlotData(energy_array, title="energy")
        default_forces = Config.get("widget", "forces_data", ForcesMode.Raw)
        if force_array.size != 0 and default_forces == ForcesMode.Norm:

            force_array = parse_array_by_atomnum(force_array, self.atoms_num_list, map_func=np.linalg.norm, axis=0)

            self._force_dataset = NepPlotData(force_array, title="force")
        else:
            self._force_dataset = NepPlotData(force_array, group_list=self.atoms_num_list, title="force")

        if float(nep_in.get("lambda_v", 1)) != 0:
            self._stress_dataset = NepPlotData(stress_array, title="stress")

            self._virial_dataset = NepPlotData(virial_array, title="virial")
        else:
            self._stress_dataset = NepPlotData([], title="stress")

            self._virial_dataset = NepPlotData([], title="virial")
    def _should_recalculate(self, nep_in: dict) -> bool:
        """判断是否需要重新计算 NEP 数据。"""
        output_files_exist = all([
            self.energy_out_path.exists(),
            self.force_out_path.exists(),
            self.stress_out_path.exists(),
            self.virial_out_path.exists()
        ])
        return not check_fullbatch(nep_in, len(self.atoms_num_list)) or not output_files_exist

    def _save_energy_data(self, potentials:npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:

        """保存能量数据到文件。"""

        try:
            ref_energies = np.array([s.per_atom_energy for s in self.structure.now_data], dtype=np.float32)

            if potentials.size  == 0:
                #计算失败 空数组
                energy_array = np.column_stack([ref_energies, ref_energies])
            else:
                energy_array = np.column_stack([potentials / self.atoms_num_list, ref_energies])
        except Exception:
            # logger.debug(traceback.format_exc())
            if potentials.size == 0:
                # 计算失败 空数组
                energy_array = np.column_stack([potentials, potentials])
            else:
                energy_array = np.column_stack([potentials / self.atoms_num_list, potentials / self.atoms_num_list])
        energy_array = energy_array.astype(np.float32)
        if energy_array.size != 0:
            np.savetxt(self.energy_out_path, energy_array, fmt='%10.8f')
        return energy_array

    def _save_force_data(self, forces: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        """保存力数据到文件。"""
        try:
            ref_forces = np.vstack([s.forces for s in self.structure.now_data], dtype=np.float32)

            if forces.size == 0:
                # 计算失败 空数组
                forces_array = np.column_stack([ref_forces, ref_forces])

            else:
                forces_array = np.column_stack([forces, ref_forces])
        except KeyError:
            MessageManager.send_warning_message("use nep3 calculator to calculate forces replace the original forces")
            forces_array = np.column_stack([forces, forces])

        except Exception:
            # logger.debug(traceback.format_exc())
            forces_array = np.column_stack([forces, forces])
            MessageManager.send_error_message("an error occurred while calculating forces. Please check the input file.")
        if forces_array.size != 0:
            np.savetxt(self.force_out_path, forces_array, fmt='%10.8f')


        return forces_array



    def _save_virial_and_stress_data(self, virials: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        """保存维里和应力数据到文件。"""
        coefficient = (self.atoms_num_list / np.array([s.volume for s in self.structure.now_data]))[:, np.newaxis]
        try:
            ref_virials = np.vstack([s.nep_virial for s in self.structure.now_data], dtype=np.float32)
            if virials.size == 0:
                # 计算失败 空数组
                virials_array = np.column_stack([ref_virials, ref_virials])
            else:
                virials_array = np.column_stack([virials, ref_virials])
        except ValueError:
            MessageManager.send_warning_message("use nep3 calculator to calculate virial replace the original virial")
            virials_array = np.column_stack([virials, virials])

        except Exception:
            MessageManager.send_error_message(f"An error occurred while calculating virial and stress. Please check the input file.")
            # logger.debug(traceback.format_exc())
            virials_array = np.column_stack([virials, virials])

        stress_array = virials_array * coefficient  * 160.21766208  # 单位转换\

        stress_array = stress_array.astype(np.float32)
        if virials_array.size != 0:
            np.savetxt(self.virial_out_path, virials_array, fmt='%10.8f')
        if stress_array.size != 0:
            np.savetxt(self.stress_out_path, stress_array, fmt='%10.8f')


        return virials_array, stress_array

    def _recalculate_and_save(self ):

        try:
            nep_potentials_array, nep_forces_array, nep_virials_array=   self.nep_calc.calculate(self.structure.now_data.tolist())
            # nep_potentials_array, nep_forces_array, nep_virials_array=self.nep_calc_thread.func_result


            # nep_potentials_array, nep_forces_array, nep_virials_array = run_nep3_calculator_process(
            #     self.nep_txt_path.as_posix(),
            #     self.structure.now_data,"calculate")
            if nep_potentials_array.size == 0:
                MessageManager.send_warning_message("The nep calculator fails to calculate the potentials, use the original potentials instead.")


            energy_array = self._save_energy_data(nep_potentials_array)
            force_array = self._save_force_data(nep_forces_array)
            virial_array, stress_array = self._save_virial_and_stress_data(nep_virials_array[:, [0, 4, 8, 1, 5, 6]])


            self.write_prediction()
            return energy_array,force_array,virial_array, stress_array
        except Exception as e:
            # logger.debug(traceback.format_exc())
            MessageManager.send_error_message(f"An error occurred while running NEP3 calculator: {e}")
            return np.array([]), np.array([]), np.array([]), np.array([])









class NepPolarizabilityResultData(ResultData):
    _polarizability_diagonal_dataset: NepPlotData
    _polarizability_no_diagonal_dataset: NepPlotData

    def __init__(self,
                 nep_txt_path: Path|str,
                 data_xyz_path: Path|str,
                 polarizability_out_path: Path|str,
                 descriptor_path: Path|str
                 ):
        super().__init__(nep_txt_path,data_xyz_path,descriptor_path)
        self.polarizability_out_path = Path(polarizability_out_path)
        # self.nep_calc = NepCalculator(model_file=self.nep_txt_path.as_posix(),
        #                               backend=NepBackend.CPU,
        #                               batch_size=Config.getint("nep", "gpu_batch_size", 1000)
        #                               )
    @property
    def datasets(self):

        return [self.polarizability_diagonal,self.polarizability_no_diagonal, self.descriptor]



    @property
    def polarizability_diagonal(self):
        return self._polarizability_diagonal_dataset
    @property
    def polarizability_no_diagonal(self):
        return self._polarizability_no_diagonal_dataset

    @property
    def descriptor(self):
        return self._descriptor_dataset

    @classmethod
    def from_path(cls, path, *, structures: list[Structure] | None = None ):
        dataset_path = Path(path)
        file_name = dataset_path.stem
        nep_txt_path = dataset_path.with_name(f"nep.txt")
        polarizability_out_path = dataset_path.with_name(f"polarizability_{file_name}.out")
        if file_name == "train":
            descriptor_path = dataset_path.with_name(f"descriptor.out")
        else:
            descriptor_path = dataset_path.with_name(f"descriptor_{file_name}.out")

        inst = cls(nep_txt_path, dataset_path, polarizability_out_path, descriptor_path)
        if structures is not None:
            try:
                inst.set_structures(structures)
            except Exception:
                pass
        return inst
    def _should_recalculate(self, nep_in: dict) -> bool:
        """判断是否需要重新计算 NEP 数据。"""
        output_files_exist = all([
            self.polarizability_out_path.exists(),

        ])
        return not check_fullbatch(nep_in, len(self.atoms_num_list)) or not output_files_exist

    def _recalculate_and_save(self ):

        try:
            # nep_polarizability_array = run_nep3_calculator_process(self.nep_txt_path.as_posix(),
            #                                                        self.structure.now_data, "polarizability")
            nep_polarizability_array = self.nep_calc.get_structures_polarizability(self.structure.now_data.tolist())


            # nep_polarizability_array=self.nep_calc_thread.func_result
            if nep_polarizability_array.size == 0:
                MessageManager.send_warning_message("The nep calculator fails to calculate the polarizability, use the original polarizability instead.")
            nep_polarizability_array = self._save_polarizability_data(  nep_polarizability_array)
            self.write_prediction()

        except Exception as e:
            # logger.debug(traceback.format_exc())
            MessageManager.send_error_message(f"An error occurred while running NEP3 calculator: {e}")

            nep_polarizability_array = np.array([])
        return nep_polarizability_array
    def _save_polarizability_data(self, polarizability: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        """保存polarizability数据到文件。"""
        nep_polarizability_array = polarizability / (self.atoms_num_list[:, np.newaxis])

        try:
            ref_polarizability = np.vstack([s.nep_polarizability for s in self.structure.now_data], dtype=np.float32)
            if polarizability.size == 0:
                # 计算失败 空数组
                polarizability_array = np.column_stack([ref_polarizability, ref_polarizability])
            else:

                polarizability_array = np.column_stack([nep_polarizability_array,
                                                        ref_polarizability

                                                        ])

        except Exception:
            # logger.debug(traceback.format_exc())
            polarizability_array = np.column_stack([polarizability, polarizability])
        polarizability_array = polarizability_array.astype(np.float32)
        if polarizability_array.size != 0:
            np.savetxt(self.polarizability_out_path, polarizability_array, fmt='%10.8f')

        return polarizability_array

    def _load_dataset(self) -> None:
        """加载或计算 NEP 数据集，并更新内部数据集属性。"""
        nep_in = read_nep_in(self.data_xyz_path.with_name("nep.in"))
        if self._should_recalculate(nep_in):
            polarizability_array = self._recalculate_and_save( )
        else:
            polarizability_array= read_nep_out_file(self.polarizability_out_path, dtype=np.float32,ndmin=2)
            if polarizability_array.shape[0]!=self.atoms_num_list.shape[0]:
                self.polarizability_out_path.unlink()
                return self._load_dataset()
        self._polarizability_diagonal_dataset = NepPlotData(polarizability_array[:, [0,1,2,6,7,8]], title="Polar Diag")

        self._polarizability_no_diagonal_dataset = NepPlotData(polarizability_array[:, [3,4,5,9,10,11]], title="Polar NoDiag")


class NepDipoleResultData(ResultData):
    _dipole_dataset: NepPlotData
    def __init__(self,
                 nep_txt_path: Path|str,
                 data_xyz_path: Path|str,
                 dipole_out_path: Path|str,
                 descriptor_path: Path|str
                 ):
        super().__init__(nep_txt_path, data_xyz_path, descriptor_path)

        self.dipole_out_path = Path(dipole_out_path)
        # self.nep_calc = NepCalculator(model_file=self.nep_txt_path.as_posix(),
        #                      backend=NepBackend.CPU,
        #                      batch_size=Config.getint("nep", "gpu_batch_size", 1000)
        #                      )
    @property
    def datasets(self):
        return [self.dipole , self.descriptor]

    @property
    def dipole(self):
        return self._dipole_dataset

    @property
    def descriptor(self):
        return self._descriptor_dataset

    @classmethod
    def from_path(cls, path, *, structures: list[Structure] | None = None ):
        dataset_path = Path(path)
        file_name = dataset_path.stem
        nep_txt_path = dataset_path.with_name(f"nep.txt")
        polarizability_out_path = dataset_path.with_name(f"dipole_{file_name}.out")

        if file_name == "train":

            descriptor_path = dataset_path.with_name(f"descriptor.out")
        else:
            descriptor_path = dataset_path.with_name(f"descriptor_{file_name}.out")

        inst = cls(nep_txt_path, dataset_path, polarizability_out_path, descriptor_path)
        if structures is not None:
            try:
                inst.set_structures(structures)
            except Exception:
                pass
        return inst


    def _should_recalculate(self, nep_in: dict) -> bool:
        """判断是否需要重新计算 NEP 数据。"""


        output_files_exist = all([
            self.dipole_out_path.exists(),

        ])
        return not check_fullbatch(nep_in, len(self.atoms_num_list)) or not output_files_exist

    def _recalculate_and_save(self ):

        try:
            # nep_dipole_array = run_nep3_calculator_process(self.nep_txt_path.as_posix(),
            #                                                self.structure.now_data, "dipole")
            nep_dipole_array = self.nep_calc.get_structures_dipole(self.structure.now_data.tolist())


            # nep_dipole_array=self.nep_calc_thread.func_result

            if nep_dipole_array.size == 0:
                MessageManager.send_warning_message("The nep calculator fails to calculate the dipole, use the original dipole instead.")
            nep_dipole_array = self._save_dipole_data(  nep_dipole_array)
            self.write_prediction()

        except Exception as e:
            # logger.debug(traceback.format_exc())
            MessageManager.send_error_message(f"An error occurred while running NEP3 calculator: {e}")

            nep_dipole_array = np.array([])
        return nep_dipole_array
    def _save_dipole_data(self, dipole: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        """保存dipole数据到文件。"""
        nep_dipole_array = dipole / (self.atoms_num_list[:, np.newaxis])

        try:
            ref_dipole = np.vstack([s.nep_dipole for s in self.structure.now_data], dtype=np.float32)
            if dipole.size == 0:
                # 计算失败 空数组
                dipole_array = np.column_stack([ref_dipole, ref_dipole])
            else:
                dipole_array = np.column_stack([nep_dipole_array,
                                            ref_dipole

                                                    ])

        except Exception:
            # logger.debug(traceback.format_exc())
            dipole_array = np.column_stack([nep_dipole_array, nep_dipole_array])
        dipole_array = dipole_array.astype(np.float32)
        if dipole_array.size != 0:
            np.savetxt(self.dipole_out_path, dipole_array, fmt='%10.8f')

        return dipole_array

    def _load_dataset(self) -> None:
        """加载或计算 NEP 数据集，并更新内部数据集属性。"""
        nep_in = read_nep_in(self.data_xyz_path.with_name("nep.in"))
        if self._should_recalculate(nep_in):
            dipole_array = self._recalculate_and_save( )
        else:
            dipole_array= read_nep_out_file(self.dipole_out_path, dtype=np.float32,ndmin=2)
            if dipole_array.shape[0]!=self.atoms_num_list.shape[0]:
                self.dipole_out_path.unlink()
                return self._load_dataset()
        self._dipole_dataset = NepPlotData(dipole_array, title="dipole")






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2025/6/18 13:21
# @Author  : 兵
# @email    : 1747193328@qq.com
from itertools import combinations

import numpy as np
from PySide6.QtWidgets import QFrame, QGridLayout
from qfluentwidgets import BodyLabel, ComboBox, ToolTipFilter, ToolTipPosition, CheckBox, EditableComboBox

from NepTrainKit import utils
from NepTrainKit.core import CardManager, process_organic_clusters, get_clusters
from NepTrainKit.custom_widget import SpinBoxUnitInputFrame
from NepTrainKit.custom_widget.card_widget import MakeDataCard
@CardManager.register_card

class CellStrainCard(MakeDataCard):
    group = "Lattice"

    card_name= "Lattice Strain"
    menu_icon=r":/images/src/images/scaling.svg"
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Make Cell Strain")

        self.init_ui()

    def init_ui(self):
        self.setObjectName("cell_strain_card_widget")


        self.engine_label=BodyLabel("Axes:",self.setting_widget)
        self.engine_type_combo=EditableComboBox(self.setting_widget)
        axes_type=["uniaxial","biaxial","triaxial","isotropic"]
        self.engine_type_combo.addItems(axes_type)
        self.engine_label.setToolTip('Pull down to select or enter a specific axis, such as X or XY')
        self.engine_label.installEventFilter(ToolTipFilter(self.engine_label, 300, ToolTipPosition.TOP))

        self.optional_frame=QFrame(self.setting_widget)
        self.optional_frame_layout = QGridLayout(self.optional_frame)
        self.optional_frame_layout.setContentsMargins(0,0,0,0)
        self.optional_frame_layout.setSpacing(2)

        self.optional_label=BodyLabel("Optional",self.setting_widget)
        self.organic_checkbox=CheckBox("Identify organic", self.setting_widget)
        self.organic_checkbox.setChecked(False)
        self.organic_checkbox.setToolTip("Treat organic molecules as rigid units")
        self.organic_checkbox.installEventFilter(ToolTipFilter(self.organic_checkbox, 300, ToolTipPosition.TOP))


        self.optional_frame_layout.addWidget(self.organic_checkbox,0,0,1,1)

        self.strain_x_label=BodyLabel("X:",self.setting_widget)
        self.strain_x_frame = SpinBoxUnitInputFrame(self)
        self.strain_x_frame.set_input(["-","% step:","%"],3,"float")
        self.strain_x_frame.setRange(-100,100)

        self.strain_x_frame.set_input_value([-5,5,1])
        self.strain_x_label.setToolTip("X-axis strain range")
        self.strain_x_label.installEventFilter(ToolTipFilter(self.strain_x_label, 300, ToolTipPosition.TOP))

        self.strain_y_label=BodyLabel("Y:",self.setting_widget)
        self.strain_y_frame = SpinBoxUnitInputFrame(self)
        self.strain_y_frame.set_input(["-","% step:","%"],3,"float")
        self.strain_y_frame.setRange(-100,100)
        self.strain_y_frame.set_input_value([-5,5,1])
        self.strain_y_label.setToolTip("Y-axis strain range")
        self.strain_y_label.installEventFilter(ToolTipFilter(self.strain_y_label, 300, ToolTipPosition.TOP))

        self.strain_z_label=BodyLabel("Z:",self.setting_widget)
        self.strain_z_frame = SpinBoxUnitInputFrame(self)
        self.strain_z_frame.set_input(["-","% step:","%"],3,"float")
        self.strain_z_frame.setRange(-100,100)
        self.strain_z_frame.set_input_value([-5,5,1])
        self.strain_z_label.setToolTip("Z-axis strain range")
        self.strain_z_label.installEventFilter(ToolTipFilter(self.strain_z_label, 300, ToolTipPosition.TOP))

        self.settingLayout.addWidget(self.engine_label,0, 0,1, 1)
        self.settingLayout.addWidget(self.engine_type_combo,0, 1, 1, 2)
        self.settingLayout.addWidget(self.optional_label, 1, 0, 1, 1)
        self.settingLayout.addWidget(self.optional_frame, 1, 1, 1,1)
        self.settingLayout.addWidget(self.strain_x_label, 2, 0, 1, 1)
        self.settingLayout.addWidget(self.strain_x_frame, 2, 1, 1,1)
        self.settingLayout.addWidget(self.strain_y_label, 3, 0, 1, 1)
        self.settingLayout.addWidget(self.strain_y_frame, 3, 1, 1,1)
        self.settingLayout.addWidget(self.strain_z_label, 4, 0, 1, 1)
        self.settingLayout.addWidget(self.strain_z_frame, 4, 1, 1,1)

    def process_structure(self, structure):
        structure_list=[]
        axes=self.engine_type_combo.currentText()
        x=self.strain_x_frame.get_input_value()
        y=self.strain_y_frame.get_input_value()
        z=self.strain_z_frame.get_input_value()
        identify_organic=self.organic_checkbox.isChecked()


        if identify_organic:
            clusters, is_organic_list = get_clusters(structure)
        strain_range=[
            np.arange(start=x[0],stop=x[1]+0.001,step=x[2]),
            np.arange(start=y[0], stop=y[1]+0.001, step=y[2]),
            np.arange(start=z[0], stop=z[1]+0.001, step=z[2]),
        ]
        cell = structure.get_cell()
        # Define possible axes (0: x, 1: y, 2: z)
        all_axes = [0, 1, 2]


        if axes == 'isotropic':
            for strain in strain_range[0]:
                new_structure = structure.copy()
                new_cell = cell.copy() * (1 + strain / 100)
                new_structure.set_cell(new_cell, scale_atoms=True)
                if identify_organic:
                    process_organic_clusters(structure, new_structure, clusters, is_organic_list )  # pyright:ignore

                strain_info = [f"all:{strain}%"]
                new_structure.info["Config_type"] = new_structure.info.get("Config_type", "") + f" Strain({'|'.join(strain_info)})"
                structure_list.append(new_structure)
        else:
            if axes == 'uniaxial':
                axes_combinations = [[i] for i in all_axes]
            elif axes == 'biaxial':
                axes_combinations = list(combinations(all_axes, 2))
            elif axes == 'triaxial':
                axes_combinations = [all_axes]
            else:
                axes_combinations = [["XYZ".index(i.upper()) for i in axes if i.upper() in "XYZ"]]
            for ax_comb in axes_combinations:
                if len(ax_comb) == 0:
                    continue
                strain_combinations = (np.array(np.meshgrid(*[strain_range[_] for _ in ax_comb])).T.reshape(-1, len(ax_comb)))
                for strain_vals in strain_combinations:
                    new_structure = structure.copy()
                    new_cell = cell.copy()
                    for ax_idx, strain in zip(ax_comb, strain_vals):
                        new_cell[ax_idx] *= (1 + strain / 100)
                    new_structure.set_cell(new_cell, scale_atoms=True)
                    if identify_organic:
                        process_organic_clusters(structure, new_structure, clusters, is_organic_list)  # pyright:ignore

                    strain_info = ["XYZ"[ax] + ":" + str(s) + "%" for ax, s in zip(ax_comb, strain_vals)]
                    new_structure.info["Config_type"] = new_structure.info.get("Config_type", "") + f" Strain({'|'.join(strain_info)})"
                    structure_list.append(new_structure)

        return structure_list

    def to_dict(self):
        data_dict = super().to_dict()
        data_dict['organic'] = self.organic_checkbox.isChecked()

        data_dict['engine_type'] = self.engine_type_combo.currentText()
        data_dict['x_range'] = self.strain_x_frame.get_input_value()
        data_dict['y_range'] = self.strain_y_frame.get_input_value()
        data_dict['z_range'] = self.strain_z_frame.get_input_value()

        return data_dict

    def from_dict(self, data_dict):
        super().from_dict(data_dict)

        self.organic_checkbox.setChecked(data_dict.get("organic", False))

        self.engine_type_combo.setText(data_dict['engine_type'])

        self.strain_x_frame.set_input_value(data_dict['x_range'])
        self.strain_y_frame.set_input_value(data_dict['y_range'])
        self.strain_z_frame.set_input_value(data_dict['z_range'])



